#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

import os

import numpy as np
import pytest
import torch

from tardis_em.utils.device import get_device
from tardis_em.utils.errors import TardisError
from tardis_em.utils.export_data import NumpyToAmira
from tardis_em.utils.load_data import (
    load_am,
    load_tiff,
    ImportDataFromAmira,
    load_mrc_file,
)
from tardis_em.utils.logo import TardisLogo
from tardis_em_analysis.filament_utils import sort_segment, cut_150_degree
from tardis_em_analysis.spatial_graph_utils import compare_splines_probability
from tardis_em_analysis.geometry_metrics import (
    angle_between_vectors,
    tortuosity,
    total_length,
)
from tardis_em.utils.utils import EarlyStopping


# TEST device.py
def test_check_device():
    dev = get_device("cpu")
    assert dev == torch.device("cpu")

    dev = get_device("1")

    if torch.cuda.is_available():
        assert dev == torch.device(type="cuda", index=1)
    else:
        assert dev == torch.device("cpu")


# TEST error.py and logo.py
def test_logo():
    logo = TardisLogo()
    # Test short
    logo(title="Test_pytest")

    # Test long
    logo(
        title="Test_pytest",
        text_1="xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
        + "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx",
    )


def test_error():
    TardisError(id_="20", py="tests/test_utils.py", desc="PyTest Failed!")


# TEST export fiels
def test_am_single_export():
    df = np.zeros((25, 4))
    df_line = np.linspace(0, 5, 25)
    df_line = np.round(df_line)
    df[:, 0] = df_line

    exporter = NumpyToAmira()
    exporter.export_amira(coords=df, file_dir="./test.am")

    assert os.path.isfile("./test.am")
    os.remove("./test.am")


def test_am_multi_export():
    df = np.zeros((25, 4))
    df_line = np.linspace(0, 5, 25)
    df_line = np.round(df_line)
    df[:, 0] = df_line

    df_1 = np.array(df)
    df_2 = np.array(df_1)

    exporter = NumpyToAmira()
    exporter.export_amira(coords=(df_1, df_2), file_dir="./test.am")

    assert os.path.isfile("./test.am")
    os.remove("./test.am")


def test_am_label_export():
    df = np.zeros((25, 4))
    df_line = np.linspace(0, 5, 25)
    df_line = np.round(df_line)
    df[:, 0] = df_line

    df_1 = np.array(df)
    df_2 = np.array(df_1)

    exporter = NumpyToAmira()
    exporter.export_amira(
        coords=(df_1, df_2), file_dir="./test.am", labels=["test1", "test2"]
    )

    assert os.path.isfile("./test.am")
    os.remove("./test.am")


# TEST file readers
def test_tif():
    tif, px = load_tiff(tiff="./tests/test_data/data_type/tif2D.tif")
    assert tif.shape == (64, 32)

    tif, px = load_tiff(tiff="./tests/test_data/data_type/tif3D.tif")
    assert tif.shape == (78, 64, 32)


def test_rec_mrc():
    mrc, px = load_mrc_file(mrc="./tests/test_data/data_type/mrc2D.mrc")
    assert mrc.shape == (64, 32)
    assert px == 23.2

    rec, px = load_mrc_file(mrc="./tests/test_data/data_type/rec2D.rec")
    assert rec.shape == (64, 32)
    assert px == 23.2

    mrc, px = load_mrc_file(mrc="./tests/test_data/data_type/mrc3D.mrc")
    assert mrc.shape == (32, 64, 78)
    assert px == 23.2

    rec, px = load_mrc_file(mrc="./tests/test_data/data_type/rec3D.rec")
    assert rec.shape == (32, 64, 78)
    assert px == 23.2


def test_am():
    am, px, ps, trans = load_am(am_file="./tests/test_data/data_type/am2D.am")

    assert am.shape == (64, 32)
    assert am.dtype == np.uint8
    assert px == 23.2
    assert np.all(trans == np.array((0, 0, 4640)))

    am, px, ps, trans = load_am(am_file="./tests/test_data/data_type/am3D.am")

    assert am.shape == (8, 256, 256)
    assert am.dtype == np.uint8
    assert px == 92.8


def test_am_sg():
    am = ImportDataFromAmira(
        src_am="./tests/test_data/data_type/am3D.CorrelationLines.am",
        src_img="./tests/test_data/data_type/am3D.am",
    )
    segments = am.get_segmented_points()
    assert segments.shape == (10, 4)

    point = am.get_points()
    assert point.shape == (10, 3)

    image, px = am.get_image()
    assert image.shape == (8, 256, 256)
    assert image.dtype == np.uint8

    px = am.get_pixel_size()
    assert px == 92.8


# TEST utils.py
def test_early_stop():
    er_stop = EarlyStopping()
    assert er_stop.counter == 0

    er_stop(val_loss=0.1)
    assert er_stop.counter == 0

    er_stop(val_loss=0.09)
    assert er_stop.counter == 0

    er_stop(val_loss=0.1)
    assert er_stop.counter == 1

    er_stop = EarlyStopping()
    assert er_stop.counter == 0

    er_stop(f1_score=0.1)
    assert er_stop.counter == 0

    er_stop(f1_score=0.15)
    assert er_stop.counter == 0

    er_stop(f1_score=0.1)
    assert er_stop.counter == 1


def test_compare_splines_probability():
    # Test with matching splines
    spline_tardis = np.array([[0, 0, 0], [1, 1, 1], [2, 2, 2]])
    spline_amira = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    threshold = 1
    assert (
        round(compare_splines_probability(spline_tardis, spline_amira, threshold), 2)
        == 0.67
    )

    # Test with non-matching splines
    spline_tardis = np.array([[0, 0], [1, 1], [2, 2]])
    spline_amira = np.array([[4, 4], [5, 5], [6, 6]])
    threshold = 1
    assert compare_splines_probability(spline_tardis, spline_amira, threshold) == 0.0

    # Test with matching splines and threshold set too high
    spline_tardis = np.array([[0, 0], [1, 1], [2, 2]])
    spline_amira = np.array([[1, 1], [2, 2], [3, 3]])
    threshold = 10
    assert compare_splines_probability(spline_tardis, spline_amira, threshold) == 1.0

    # Test with matching splines and threshold set too low
    spline_tardis = np.array(
        [[0, 0, 0], [1, 1, 1], [2, 2, 2], [4, 4, 4], [5, 5, 5], [6, 6, 6]]
    )
    spline_amira = np.array([[1, 1, 1], [2, 2, 2], [3, 3, 3]])
    threshold = 1
    assert (
        round(compare_splines_probability(spline_tardis, spline_amira, threshold), 2)
        == 0.33
    )

    # Test with matching splines and threshold set too low
    spline_tardis = np.array([[1, 1, 1], [2, 2, 2], [4, 4, 4]])
    spline_amira = np.array(
        [[0, 0, 0], [1, 1, 1], [2, 2, 2], [4, 4, 4], [5, 5, 5], [6, 6, 6]]
    )
    threshold = 1
    assert (
        round(compare_splines_probability(spline_tardis, spline_amira, threshold), 2)
        == 1.0
    )

    # Test with empty spline
    spline_tardis = np.array([[0, 0], [1, 1], [2, 2]])
    spline_amira = np.array([])
    threshold = 1
    assert compare_splines_probability(spline_tardis, spline_amira, threshold) == 0.0


# TEST spatial_graph_utils.py
def test_sort_segment():
    # Define an example point cloud with unsorted points
    unsorted_points = np.array([[1, 2, 3], [4, 5, 6], [2, 3, 4], [3, 4, 5]])

    # Call the sort_segment function
    sorted_points = sort_segment(unsorted_points)

    # Define what the expected output should be for the given input
    expected_output = np.array([[1, 2, 3], [2, 3, 4], [3, 4, 5], [4, 5, 6]])

    # Assert that the output from the function matches the expected output
    np.testing.assert_array_equal(sorted_points, expected_output)


def test_tortuosity():
    # Define an example set of coordinates representing a straight line
    straight_line = np.array([[0, 0, 0], [1, 0, 0], [2, 0, 0]])
    # Tortuosity of a straight line should be 1
    assert tortuosity(straight_line) == pytest.approx(1.0, abs=1e-6)

    # Define an example set of coordinates representing a curved path
    curved_path = np.array([[0, 0, 0], [1, 1, 0], [2, 0, 0]])
    # Tortuosity of a curved path should be greater than 1
    assert tortuosity(curved_path) > 1.0

    # Test with only one point (should return 1.0)
    one_point = np.array([[0, 0, 0]])
    assert tortuosity(one_point) == 1.0

    # Test with no points (should return 1.0)
    no_points = np.array([])
    assert tortuosity(no_points) == 1.0


def test_total_length():
    # Define an example set of coordinates representing a straight line
    straight_line = np.array([[0, 0, 0], [1, 0, 0], [2, 0, 0]])
    # Total length of a straight line should be equal to the distance between the first and last point
    assert total_length(straight_line) == pytest.approx(2.0, abs=1e-6)

    # Define an example set of coordinates representing a curved path
    curved_path = np.array([[0, 0, 0], [1, 1, 0], [2, 0, 0]])
    # Total length of a curved path should be greater than the distance between the first and last point
    assert total_length(curved_path) > 2.0

    # Test with only one point (should return 0.0)
    one_point = np.array([[0, 0, 0]])
    assert total_length(one_point) == 0.0


def test_angle_between_vectors():
    # Test with perpendicular vectors
    v1 = np.array([1, 0, 0])
    v2 = np.array([0, 1, 0])
    assert angle_between_vectors(v1, v2) == 90.0

    # Test with parallel vectors
    v1 = np.array([1, 0, 0])
    v2 = np.array([2, 0, 0])
    assert angle_between_vectors(v1, v2) == 0.0

    # Test with anti-parallel vectors
    v1 = np.array([1, 0, 0])
    v2 = np.array([-1, 0, 0])
    assert angle_between_vectors(v1, v2) == 180.0

    # Test with vectors at 45 degrees
    v1 = np.array([1, 0, 0])
    v2 = np.array([1, 1, 0])
    assert angle_between_vectors(v1, v2) == pytest.approx(45.0, abs=1e-6)

    # Test with negative values
    v1 = np.array([-1, 0, 0])
    v2 = np.array([0, -1, 0])
    assert angle_between_vectors(v1, v2) == 90.0


def test_cut_150_degree():
    # Define a segment with a sharp angle that should be cut
    segments = np.array(
        [
            [0, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 1, 1, 0],
            [0, 2, 1, 0],
        ]
    )

    was_cut, cut_segments = cut_150_degree(segments)
    assert was_cut
    assert cut_segments.shape == (4, 4)
    assert np.all(cut_segments[0, :] == [0, 0, 0, 0])
    assert np.all(cut_segments[1, :] == [0, 1, 0, 0])
    assert np.all(cut_segments[2, :] == [1, 1, 1, 0])
    assert np.all(cut_segments[3, :] == [1, 2, 1, 0])

    # Define a segment with all angles greater than 150 degrees
    segments = np.array(
        [
            [0, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 2, 0, 0],
            [0, 3, 0, 0],
        ]
    )

    was_cut, cut_segments = cut_150_degree(segments)
    assert not was_cut
    assert cut_segments.shape == segments.shape
    assert np.all(cut_segments == segments)



# #####################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
# #####################################################################

import numpy as np

from tardis_em.cnn.data_processing.interpolation import interpolation


def test_interpolation_spline_x():
    points = np.array([[1, 2, 3], [4, 2, 3]])
    expected_result = np.array([[1, 2, 3], [2, 2, 3], [3, 2, 3], [4, 2, 3]])
    assert np.array_equal(interpolation(points), expected_result)


def test_interpolation_spline_y():
    points = np.array([[1, 2, 3], [1, 4, 3]])
    expected_result = np.array([[1, 2, 3], [1, 3, 3], [1, 4, 3]])
    assert np.array_equal(interpolation(points), expected_result)


def test_interpolation_spline_z():
    points = np.array([[1, 2, 3], [1, 2, 4]])
    expected_result = np.array([[1, 2, 3], [1, 2, 4]])
    assert np.array_equal(interpolation(points), expected_result)



# #####################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
# #####################################################################

import os
import subprocess
import subprocess as subp
import sys

from tardis_em.utils.errors import TardisError
from tardis_em.utils.logo import TardisLogo
from tardis_em._version import version


def env_exists(env_name: str) -> bool:
    """
    Simple check if environment exist

    Args:
        env_name (str): Environment name

    Returns:
        bool: If True, environment exist.
    """
    command = "conda info --envs"
    process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    envs = output.decode().split("\n")

    return env_name in envs


def py(python: str):
    """
    Run pytest on specified python version inside conda environment

    Args:
        python (str): Python version.

    Returns:
        list: Output log from pytest.
    """
    os.chdir("../../")  # move to root dir

    if env_exists("PythonEnvTest"):
        # Remove test environment
        subp.run("conda remove -n PythonEnvTest --all -y", shell=True)
    else:
        # Create clean environment
        subp.run("conda create --name PythonEnvTest -y", shell=True)

    # Set up Python 3.X env and update
    subp.run(
        f"conda run -n PythonEnvTest conda install python={'3.' + python[1:]} -y",
        shell=True,
    )

    # Check and reinstall if needed requirements
    subp.run(
        "conda run -n PythonEnvTest pip install -r requirements-dev.txt", shell=True
    )
    subp.run("conda run -n PythonEnvTest pip install -r requirements.txt", shell=True)

    # Clean-up
    subp.run("conda run -n PythonEnvTest conda clean -a -y", shell=True)
    subp.run("conda run -n PythonEnvTest pip cache purge", shell=True)

    # Install tardis_em-pytorch
    subp.run("conda run -n PythonEnvTest pip install -e .", shell=True)

    # Test on Python 3.X.*
    return subp.run(
        "conda run -n PythonEnvTest pytest", shell=True, capture_output=True
    )


if __name__ == "__main__":
    tardis_progress = TardisLogo()
    tardis_progress(title=f"Development - TARDIS {version} - pytest")

    """ Run Pytest on python 3.7 - 3.11"""
    if sys.platform != "darwin":  # Python 3.7 on macOS is only available throw x64
        out = py(python="37")
        if not out.retuncode == 0:
            TardisError("20", f"{out}" "Pyton 3.7 pytest Failed")
            exit()

    out = py(python="38")
    if not out.returncode == 0:
        TardisError("20", f"{out}" "Pyton 3.8 pytest Failed")
        exit()

    out = py(python="39")
    if not out.returncode == 0:
        TardisError("20", f"{out}" "Pyton 3.9 pytest Failed")
        exit()

    out = py(python="310")
    if not out.returncode == 0:
        TardisError(f"{out}" "Pyton 3.10 pytest Failed")
        exit()

    # !!! Python 3.11 missing compatibility with numpy open3d and pytorch !!!
    #
    # out = py(python='311')
    # if not out.returncode == 0:
    #     TardisError(f'{out}'
    #                 'Pyton 3.11 pytest Failed')
    #     exit()

    """ Return output """
    tardis_progress(
        title=f"Development - TARDIS {version} - pytest",
        text_1="All test passed correctly on python 3.7, 3.8, 3.9, 3.10",
        text_2="Sphinx-build Completed.",
    )



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
from typing import Union

import torch
from torch import nn

from tardis_em.dist_pytorch.sparse_model.modules import (
    SparsTriangularUpdate,
)
from tardis_em.dist_pytorch.model.modules import GeluFeedForward


class SparseDistStack(nn.Module):
    """
    Represents a stack of SparseDistLayer modules. Each layer in the stack processes input data
    based on pairwise features and nearest neighbor interactions. The stack enables cascading
    computations in multiple layers to enhance the representation of the input.

    This class is designed for processing pairwise relationships with sparse data structures,
    particularly in machine learning frameworks where modular and scalable architecture is needed.
    It allows easy modification of the number of layers and their properties.
    """

    def __init__(self, pairs_dim: int, num_layers=1, ff_factor=4, knn=8):
        """
        Initializes the class with specified parameters and creates a module list containing
        multiple SparseDistLayer instances. Each layer is initialized with the specified
        parameters for pairs dimension, feed-forward factor, and k-nearest neighbors.

        :param pairs_dim: The dimension of the input pairwise data.
        :type pairs_dim: int
        :param num_layers: The number of layers to instantiate. Defaults to 1.
        :type num_layers: int, optional
        :param ff_factor: The multiplication factor for the feed-forward layer. Defaults to 4.
        :type ff_factor: int, optional
        :param knn: The number of k-nearest neighbors to use. Defaults to 8.
        :type knn: int, optional
        """
        super().__init__()
        self.layers = nn.ModuleList()
        for _ in range(num_layers):
            self.layers.append(
                SparseDistLayer(pairs_dim=pairs_dim, ff_factor=ff_factor, knn=knn)
            )

    def __len__(self) -> int:
        """
        Returns the number of layers in the stack.
        """
        return len(self.layers)

    def __getitem__(self, i: int):
        """
        Retrieves the layer at the given index from the collection of layers.

        This method provides access to a specific layer object, based on the provided
        index. The index corresponds to the position of the layer within the internal
        list. If the index is out of the range of the collection, an exception will be
        raised.

        :param i: The index of the layer to retrieve.
        :type i: int
        :return: The layer object at the specified index.
        :rtype: Any
        """
        return self.layers[i]

    def forward(
        self, edge_features: torch.tensor, indices: list
    ) -> Union[torch.tensor, list]:
        """
        Processes the given edge features and indices through a sequence of layers, updating the edge features
        at each step by passing them through the corresponding layer along with the provided indices. The
        resulting edge features after processing through all layers are returned.

        :param edge_features: A tensor containing the features associated with the edges. Each layer uses this
            tensor during its processing step to compute the updated edge features.
        :param indices: A list of indices corresponding to the edges. These indices are used during the
            computation in each layer when processing the edge features.
        :return: A tensor containing the updated edge features after processing through all layers sequentially.
        """
        for layer in self.layers:
            edge_features = layer(h_pairs=edge_features, indices=indices)

        return edge_features


class SparseDistLayer(nn.Module):
    """
    SparseDistLayer is a neural network module designed for updating sparse edge
    features, leveraging nearest neighbors, feed-forward networks, and triangular
    updates. The main purpose of this layer is to process and transform sparse
    relationships while effectively handling computational efficiencies for
    large-scale data representations in sparse tensor formats.

    This class serves as an essential component in advanced neural network
    architectures and is particularly tailored for sparse data processing,
    supporting updates for both edges and potentially nodes in future iterations.
    """

    def __init__(self, pairs_dim: int, ff_factor=4, knn=8):
        """
        Initializes an object with configuration for edge updates and feed-forward normalization.

        :param pairs_dim: The dimensionality of input pairs.
        :param ff_factor: The factor determining the hidden layer dimensionality in the
                          feed-forward network. Defaults to 4.
        :param knn: The number of nearest neighbors for sparse triangular updates. Defaults to 8.
        """
        super().__init__()
        self.pairs_dim = pairs_dim
        self.channel_dim = int(pairs_dim / 4)
        if self.channel_dim <= 0:
            self.channel_dim = 1

        # ToDo Node optional features update

        # ToDO Edge optional MHA update

        # Edge triangular update
        self.row_update = SparsTriangularUpdate(
            input_dim=pairs_dim, channel_dim=self.channel_dim, axis=1, knn=knn
        )
        self.col_update = SparsTriangularUpdate(
            input_dim=pairs_dim, channel_dim=self.channel_dim, axis=0, knn=knn
        )

        # Edge GeLu FFN normalization layer
        self.pair_ffn = GeluFeedForward(
            input_dim=pairs_dim, ff_dim=pairs_dim * ff_factor
        )

    def update_nodes(self) -> torch.tensor:
        """
        Placeholder method for future node update functionality.
        """
        pass

    def update_edges(
        self, h_pairs: torch.tensor, indices: list
    ) -> Union[torch.tensor, list]:
        """
        Updates the edge features by applying transformations to the input node
        feature pair tensor. These transformations include row-wise, column-wise
        updates, and a feed-forward network applied on the node pairs. The method
        returns the modified tensor, representing updated edge features.

        :param h_pairs: Tensor representing the features of node pairs that
            need to be updated.
        :param indices: List of indices used for performing updates on the
            edge features.
        :return: Tensor representing the updated edge features, which includes
            the applied transformations.
        :rtype: Union[torch.tensor, list]
        """
        # ToDo Convert node features to edge shape

        # Update edge features
        h_pairs = (
            h_pairs
            + self.row_update(x=h_pairs, indices=indices)
            + self.col_update(x=h_pairs, indices=indices)
        )

        return h_pairs + self.pair_ffn(x=h_pairs)

    def forward(
        self, h_pairs: torch.tensor, indices: list
    ) -> Union[torch.tensor, list]:
        """
        Updates the features of input nodes and reshapes them to have the required edge
        shape. This method first updates the edge features based on the given pairs and
        indices, then returns the updated features.

        :param h_pairs: A tensor representing the input node features. Dimensions and
            meaning depend on the tensor's context within the model.
        :param indices: A list of indices used to update the edge features. The indices
            specify how the input tensor `h_pairs` is altered.
        :return: A tensor or list with updated edge features, containing the transformed
            input node values based on the specified indices. The return type is dependent
            on the implementation of `update_edges`.
        """
        # ToDo Update node features and convert to edge shape

        # Update edge features
        h_pairs = self.update_edges(h_pairs=h_pairs, indices=indices)

        return h_pairs



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
import json
import sys
import time
from datetime import datetime
from os import listdir, getcwd, mkdir
from os.path import isdir, isfile, join, expanduser
from typing import Optional, Union
import platform

import numpy as np
import pandas as pd
import tifffile.tifffile as tif
import torch

from tardis_em.dist_pytorch.utils.utils import VoxelDownSampling

from tardis_em.dist_pytorch.datasets.patches import PatchDataSet
from tardis_em.dist_pytorch.dist import build_dist_network
from tardis_em.dist_pytorch.utils.build_point_cloud import BuildPointCloud
from tardis_em.dist_pytorch.utils.segment_point_cloud import PropGreedyGraphCut
from tardis_em.cnn.data_processing.draw_mask import (
    draw_semantic_membrane,
    draw_instances,
)
from tardis_em.cnn.data_processing.stitch import StitchImages
from tardis_em.cnn.data_processing.trim import trim_with_stride
from tardis_em.cnn.datasets.dataloader import PredictionDataset
from tardis_em.cnn.cnn import build_cnn_network
from tardis_em.cnn.data_processing.scaling import scale_image
from tardis_em.utils.aws import get_weights_aws, get_all_version_aws
from tardis_em.utils.device import get_device
from tardis_em.utils.errors import TardisError
from tardis_em.utils.export_data import NumpyToAmira, to_am, to_mrc, to_stl
from tardis_em.utils.load_data import load_am, ImportDataFromAmira, load_image
from tardis_em.utils.logo import print_progress_bar, TardisLogo
from tardis_em.utils.normalization import (
    MeanStdNormalize,
    RescaleNormalize,
    adaptive_threshold,
)
from tardis_em.utils.setup_envir import build_temp_dir, clean_up
from tardis_em_analysis.spatial_graph_utils import (
    FilterSpatialGraph,
    FilterConnectedNearSegments,
    SpatialGraphCompare,
    ComputeConfidenceScore,
)
from tardis_em_analysis.filament_utils import sort_by_length, resample_filament
from tardis_em_analysis.geometry_metrics import length_list
from tardis_em._version import version

# try:
#     from tardis_em.utils.ota_update import ota_update
#
#     ota = ota_update(status=True)
# except ImportError:
#     ota = ""
ota = ""

# Pytorch CUDA optimization
if torch.cuda.is_available():
    torch.backends.cudnn.benchmark = True


class GeneralPredictor:
    """
    Summary of what the class does.

    The GeneralPredictor class is designed for handling image-based predictions
    using various neural network architectures. It provides functionalities
    to set up the environment, preprocess data, run predictions with or
    without rotation, and post-process the results for further usage or
    analysis. This class also includes methods to handle different file formats,
    apply various normalization techniques, and output results in specified formats.
    The class is modular enough to work with instances, semantic predictions, and
    custom settings based on the input parameters and configurations.
    """

    def __init__(
        self,
        predict: str,
        dir_s: Union[str, tuple[np.ndarray], np.ndarray],
        binary_mask: bool,
        output_format: str,
        patch_size: int,
        convolution_nn: str,
        cnn_threshold: str,
        dist_threshold: float,
        points_in_patch: int,
        predict_with_rotation: bool,
        instances: bool,
        device_s: str,
        debug: bool,
        checkpoint: Optional[list] = None,
        model_version: Optional[int] = None,
        correct_px: float = None,
        normalize_px: float = None,
        amira_prefix: str = None,
        filter_by_length: int = None,
        connect_splines: int = None,
        connect_cylinder: int = None,
        amira_compare_distance: int = None,
        amira_inter_probability: float = None,
        tardis_logo: bool = True,
        continue_b: bool = False,
    ):
        """
        This class initializes the configuration and parameters required for predictive models
        based on convolutional neural networks and builds the necessary handlers to process,
        transform, and analyze the provided data and predictions. It handles pre-processing,
        prediction, and post-processing steps, including normalization, stitching, spatial graph
        comparison, and various filters.

        :param predict: The predictive model type to be used.
        :param dir_s: The directory path or dataset array for input data processing.
        :param binary_mask: Flag indicating if a binary mask will be used.
        :param output_format: Specifies the output format for results.
        :param patch_size: Size of the patches in the image dataset.
        :param convolution_nn: Type of convolution neural network to use.
        :param cnn_threshold: Threshold value for CNN predictions.
        :param dist_threshold: Distance threshold used in graph segmentation.
        :param points_in_patch: Maximum number of points to include in a patch.
        :param predict_with_rotation: Flag to determine if prediction should consider rotations.
        :param instances: Flag to enable instance-based predictions.
        :param device_s: The computation device (CPU/GPU) to be used.
        :param debug: Flag to enable debugging mode.
        :param checkpoint: Optional, list of checkpoints for loading model weights.
        :param model_version: Optional, model version identifier.
        :param correct_px: Optional, value to correct pixel dimensions if needed.
        :param normalize_px: Optional, value to normalize pixel dimensions if required.
        :param amira_prefix: Optional, prefix for Amira file formats.
        :param filter_by_length: Optional, specifies filter to exclude segments of short lengths.
        :param connect_splines: Optional, parameter for connecting splines at close distances.
        :param connect_cylinder: Optional, radius for connecting cylindrical regions.
        :param amira_compare_distance: Optional, distance threshold for spatial graph comparison.
        :param amira_inter_probability: Optional, interaction threshold for Amira spatial graph probabilities.
        :param tardis_logo: Flag to handle tardis logo processing.
        :param continue_b: Flag to indicate if processing should resume from previous state.

        :raises: Raises errors for various invalid configurations or unsupported operation parameters.
        """
        self.continue_b = continue_b
        self.transformation, self.px, self.image = None, None, None
        self.tardis_logo = tardis_logo
        self.tardis_progress = None
        self.title = None

        # Directories and dataset info
        self.dir = dir_s
        self.output_format = output_format
        self.predict = predict
        if self.predict in ["Membrane2D", "Microtubule_tirf"]:
            self.expect_2d = True
        else:
            self.expect_2d = False
        self.amira_prefix = amira_prefix
        self.checkpoint = checkpoint
        self.model_version = model_version
        self.correct_px = correct_px
        self.normalize_px = normalize_px

        # Pre-processing setting
        self.cnn, self.dist = None, None
        self.patch_size = patch_size
        self.points_in_patch = points_in_patch
        self.pc_hd, self.pc_ld = np.zeros((0, 3)), np.zeros((0, 3))
        self.coords_df = np.zeros((1, 3))
        self.transformation = [0, 0, 0]
        self.segments, self.segments_filter = None, None

        # Prediction setting
        self.convolution_nn = convolution_nn
        self.cnn_threshold = cnn_threshold
        self.dist_threshold = dist_threshold
        self.rotate = predict_with_rotation

        # Global flags
        self.binary_mask = binary_mask
        self.predict_instance = instances
        self.device = get_device(device_s)
        self.debug = debug
        self.semantic_header, self.instance_header, self.log_prediction = [], [], []
        self.eta_predict = "NA"

        """Initial Setup"""
        if debug:
            self.str_debug = " <Debugging Mode>"
        else:
            self.str_debug = ""

        # Check for the spatial graph in the folder from amira/tardis_em comp.
        self.amira_check = False
        if isinstance(self.dir, str):
            if isdir(join(self.dir, "amira")):
                self.amira_check = True
                self.dir_amira = join(self.dir, "amira")

        # Searching for available images for prediction
        self.available_format = (
            ".tif",
            ".tiff",
            ".mrc",
            ".rec",
            ".am",
            ".map",
            ".npy",
        )
        self.omit_format = (
            ".nd2",
            "mask.tif",
            "mask.tiff",
            "mask.mrc",
            "mask.rec",
            "Correlation_Lines.am",
            "mask.am",
            "mask.map",
        )

        """Build handler's"""
        # Build handler's for reading data to correct format
        self.normalize = RescaleNormalize(clip_range=(1, 99))  # Normalize histogram
        self.mean_std = MeanStdNormalize()  # Standardize with mean and std

        # Sigmoid whole predicted image
        self.sigmoid = torch.nn.Sigmoid()

        # Build handler's for transforming data
        self.image_stitcher = StitchImages()
        self.post_processes = BuildPointCloud()

        # self.get_file_list()

        # Build handler for DIST input and output
        if self.predict_instance:
            self.patch_pc = PatchDataSet(
                max_number_of_points=points_in_patch, graph=False
            )

            if predict in [
                "Actin",
                "Microtubule",
                "Membrane2D",
                "General_filament",
                "Microtubule_tirf",
            ]:
                self.GraphToSegment = PropGreedyGraphCut(
                    threshold=dist_threshold, connection=2, smooth=True
                )

                if predict == "Membrane2D":
                    self.filter_splines = FilterConnectedNearSegments(
                        distance_th=connect_splines,
                        cylinder_radius=connect_cylinder,
                    )
                else:
                    self.filter_splines = FilterSpatialGraph(
                        connect_seg_if_closer_then=connect_splines,
                        cylinder_radius=connect_cylinder,
                        filter_short_segments=(
                            filter_by_length if filter_by_length is not None else 0
                        ),
                    )
                self.compare_spline = SpatialGraphCompare(
                    distance_threshold=amira_compare_distance,
                    interaction_threshold=amira_inter_probability,
                )
                self.score_splines = ComputeConfidenceScore()
            elif predict in ["Membrane", "General_object"]:
                self.GraphToSegment = PropGreedyGraphCut(
                    threshold=dist_threshold, connection=8
                )

        # Build handler to output amira file
        self.create_headers()
        self.amira_file = NumpyToAmira()
        self.amira_file_points = NumpyToAmira(as_point_cloud=True)

        # Sanity error checks
        self.init_check()

        # Build NN from checkpoints
        self.build_NN(NN=self.predict)

    def create_headers(self):
        """
        Creates ASCII headers and initializes logging information for a spatial graph
        prediction process. The headers include project details, directory setup,
        neural network configurations for semantic and instance segmentation, and
        other metadata. It verifies paths, logs prediction states appropriately,
        and dynamically determines the model version depending on the predictive
        needs and available checkpoints.

        :raises: None
        :parameter self.dir: Directory path for file output or processing.
            Default is retrieved current working directory if unset.
        :parameter self.output_format: Format of the output data produced during
           prediction, e.g. as,send LabelSetsivet nn
        """
        dir_ = self.dir if isinstance(self.dir, str) else "np.ndarray"
        if dir_ == ".":
            dir_ = getcwd()

        main_ = [
            "###############################################################################",
            "# TARDIS - Transformer And Rapid Dimensionless Instance Segmentation (R)      #",
            f"# tardis_em v{version}                                                           #",
            f"# MIT License * 2021-{datetime.now().year} | Robert Kiewisz & Tristan Bepler                   #",
            "###############################################################################",
            "",
            f"Directory: {dir_}",
            f"Output Format: {self.output_format}",
            f"Predict: {self.predict}",
            f"Device: {self.device}",
            "",
        ]

        if self.model_version is None:
            model = self.predict

            if model == "Actin":
                model = "actin_3d"
            elif model == "Microtubule":
                model = "microtubules_3d"
            elif model == "Microtubule_tirf":
                model = "microtubules_tirf"
            elif model == "Membrane":
                model = "membrane_3d"
            elif model == "Membrane2D":
                model = "membrane_2d"
            else:
                model = "None"

            if model != "None":
                if self.model_version is None:
                    try:
                        model_version = get_all_version_aws(
                            self.convolution_nn, "32", model
                        )
                        model_version = model_version[-1]
                    except:
                        if isfile(
                            join(
                                expanduser("~"),
                                ".tardis_em",
                                "fnet_attn_32",
                                model,
                                "model_version.json",
                            )
                        ):
                            model_version = json.load(
                                open(
                                    join(
                                        expanduser("~"),
                                        ".tardis_em",
                                        "fnet_attn_32",
                                        model,
                                        "model_version.json",
                                    )
                                )
                            )
                            model_version = model_version["version"]
                        else:
                            model_version = "Local"
                else:
                    model_version = self.model_version
            else:
                model_version = "None"
        else:
            model_version = self.model_version

        self.semantic_header = [
            (
                f"CNN type: {self.convolution_nn} {model_version} "
                if self.checkpoint[0] is None
                else f"CNN model loaded from checkpoint: {self.checkpoint[0]}"
                f"Image patch size used for CNN: {self.patch_size}"
            ),
            f"CNN threshold: {self.cnn_threshold}",
            f"CNN predicted with 4x 90 degrees rotations: {self.rotate}",
        ]

        self.instance_header = [
            (
                "DIST 2D model"
                if self.predict
                in ["Actin", "Microtubule", "Membrane2D", "Microtubule_tirf"]
                else "DIST 3D model"
            ),
            (
                ""
                if self.checkpoint[1] is None
                else f"DIST model loaded from checkpoint: {self.checkpoint[1]}"
                f"DIST threshold: {self.dist_threshold}"
            ),
            f"Maximum number of points used in single DIST run: {self.points_in_patch}",
        ]
        init_log = (
            main_
            + ["----Semantic Segmentation----"]
            + self.semantic_header
            + ["", "----Instance Segmentation----"]
            + self.instance_header
            + [""]
        )

        if self.continue_b and isfile(
            join(self.dir, "Predictions", "prediction_log.txt")
        ):
            self.log_prediction = np.genfromtxt(
                join(self.dir, "Predictions", "prediction_log.txt"),
                delimiter=",",
                dtype=str,
            )
            self.log_prediction = [str(s) for s in self.log_prediction]
            self.log_prediction = (
                self.log_prediction + ["", "---Continue prediction---", ""] + init_log
            )
        else:
            self.log_prediction = init_log

    def init_check(self):
        """
        Perform initialization and validation checks for a segmentation prediction task.

        This method performs several checks and initializations before executing a segmentation
        task. It first validates whether the requested prediction type is supported. If support is
        enabled for a TARDIS logo, additional checks are performed for user configurations, and error
        messages are displayed if any invalid settings are detected. Subsequently, the log output
        initialization begins with a defined title according to the segmentation type and configuration.

        This function also validates the chosen output format to ensure its compatibility with the
        system (e.g., checking for unsupported machine types like ARM64 or ensuring that at least one
        valid output format is selected). Invalid configurations or unsupported settings result in
        terminating the execution with proper error messaging.

        :param self: Instance of the class to which the initialization belongs.

        :raises AssertionError: If the requested segmentation type is unsupported.
        :raises TardisError: If any invalid configuration is detected based on
                              TARDIS-specific rules (e.g., invalid output format, machine-type
                              dependencies).
        """
        msg = f"TARDIS v.{version} supports only MT and Mem segmentation!"
        assert_b = self.predict in [
            "Actin",
            "Membrane2D",
            "Membrane",
            "Microtubule",
            "Microtubule_tirf",
            "General_filament",
            "General_object",
        ]
        if self.tardis_logo:
            # Check if users ask to predict the correct structure
            if not assert_b:
                TardisError(
                    id_="01",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
        else:
            assert assert_b, msg

        # Initiate log output
        if self.tardis_logo:
            self.tardis_progress = TardisLogo()

            if isdir(join(self.dir, "amira")):
                self.title = (
                    f"Fully-automatic Instance {self.predict} segmentation module "
                    f"with Amira comparison {self.str_debug}"
                )
            elif self.predict_instance:
                if self.output_format.startswith("None"):
                    self.title = (
                        f"Fully-automatic Instance {self.predict} segmentation module | {ota} "
                        f"{self.str_debug}"
                    )
                else:
                    self.title = (
                        f"Fully-automatic Semantic-Instance {self.predict} segmentation module | {ota} "
                        f"{self.str_debug}"
                    )
            else:
                self.title = (
                    f"Fully-automatic Semantic {self.predict} segmentation module | {ota} "
                    f"{self.str_debug}"
                )

            self.tardis_progress(title=self.title, text_2=f"Device: {self.device}")

        # Check for any other errors
        # Early stop if not semantic of instance was specified
        msg = f"Require that at lest one output format is not None but {self.output_format} was given!"
        assert_b = self.output_format == "None_None"
        if assert_b:
            if self.tardis_logo:
                TardisError(
                    id_="151",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
            else:
                assert assert_b, msg

        # Check for know error if using ARM64 machine
        msg = f"STL output is not allowed on {platform.machine()} machine type!"
        assert_b = platform.machine() == "aarch64"
        if self.output_format.endswith("stl"):
            if self.tardis_logo:
                TardisError(
                    id_="151",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
            else:
                assert not assert_b, msg

    def build_NN(self, NN: str):
        """
        Builds the neural network and distance prediction modules based on the specified
        neural network (NN) type. Depending on the NN type, the appropriate configurations
        and pre-trained weights are loaded for CNN and DIST networks. This method supports
        multiple NN types, including Actin, Microtubule, Membrane, and General models.
        Additionally, configurations for 2D, 3D, and other specialized models are supported.

        :param NN: A string denoting the neural network type. Supported types include
            "Actin", "Microtubule", "Microtubule_tirf", "Membrane2D", "Membrane",
            or any type starting with "General".
        :type NN: str
        """
        if NN in ["Actin", "Microtubule", "Microtubule_tirf"]:
            # None - default value
            # 0.0 - Do not normalize
            # > 0 - normalzie to specific A resolution
            if self.normalize_px is None:
                self.normalize_px = 25
            elif self.normalize_px == 0.0:
                self.normalize_px = None

            if NN == "Actin" and self.normalize_px is not None:
                if self.normalize_px == 0.0:
                    self.normalize_px = None
                else:
                    self.normalize_px = 15

            # Build CNN network with loaded pre-trained weights
            if not self.binary_mask:
                if NN in ["Actin", "Microtubule"]:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type=(
                            "microtubules_3d" if NN == "Microtubule" else "actin_3d"
                        ),
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        device=self.device,
                    )
                else:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type="microtubules_tirf",
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        _2d=True,
                        device=self.device,
                    )
            # Build DIST network with loaded pre-trained weights
            if not self.output_format.endswith("None"):
                self.dist = Predictor(
                    checkpoint=self.checkpoint[1],
                    network="dist",
                    subtype="triang",
                    model_type="2d",
                    model_version=None,
                    device=self.device,
                )
        elif NN in ["Membrane2D", "Membrane"]:
            # None - default value
            # 0.0 - Do not normalize
            # > 0 - normalzie to specific A resolution
            if self.normalize_px is None:
                self.normalize_px = 15
            elif self.normalize_px == 0.0:
                self.normalize_px = None

            # Build CNN network with loaded pre-trained weights
            if NN == "Membrane2D":
                if not self.binary_mask:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type="membrane_2d",
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        device=self.device,
                        _2d=True,
                    )

                # Build DIST network with loaded pre-trained weights
                if not self.output_format.endswith("None"):
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="2d",
                        model_version=None,
                        device=self.device,
                    )
            else:
                if not self.binary_mask:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type="membrane_3d",
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        device=self.device,
                    )

                # Build DIST network with loaded pre-trained weights
                if not self.output_format.endswith("None"):
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="3d",
                        model_version=None,
                        device=self.device,
                    )
        elif NN.startswith("General"):
            self.cnn = Predictor(
                checkpoint=self.checkpoint[0],
                model_version=self.model_version,
                network=self.convolution_nn,
                img_size=self.patch_size,
                sigmoid=False,
                device=self.device,
                _2d=self.expect_2d,
            )
            if not self.output_format.endswith("None"):
                if NN.endswith("filament"):
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="2d",
                        model_version=None,
                        device=self.device,
                    )
                else:
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="3d",
                        model_version=None,
                        device=self.device,
                    )

    def load_data(self, id_name: Union[str, np.ndarray]):
        """
        Loads and processes image data or point cloud data from a specified file or array. Depending on the
        input type, the function determines if the data is an AmiraMesh 3D ASCII file, a general image file,
        or a preloaded array. It performs normalization, sanity checks, and prepares the data for further
        processing.

        :param id_name:
            Specifies either the path to the file to be loaded or an already loaded numpy array.
            Can be a string filename or a numpy array.

        :raises AssertionError:
            - If the Amira Spatial Graph has dimensions other than 4.
            - If the loaded image's dtype is not `float32` after normalization.
            - If the processed binary mask dtype is not `int8` or `uint8`.

        :raises TardisError:
            - If `tardis_logo` is True and any of the aforementioned conditions fail.

        :raises SystemExit:
            - If certain errors occur during processing and `tardis_logo` is True.

        :return: None
        """
        self.log_prediction.append("\n")
        self.log_prediction.append(f"Loaded image: {id_name}")

        # Build temp dir
        build_temp_dir(dir_s=self.dir)

        if isinstance(id_name, str):
            # Load image file
            if id_name.endswith(".am"):
                am = open(join(self.dir, id_name), "r", encoding="iso-8859-1").read(500)

                if "AmiraMesh 3D ASCII" in am:
                    self.amira_image = False
                    self.pc_hd = ImportDataFromAmira(
                        join(self.dir, id_name)
                    ).get_segmented_points()

                    self.image = None
                    self.px = self.correct_px
                    self.transformation = [0, 0, 0]

                    assert_b = self.pc_hd.shape[1] == 4
                    msg = f"Amira Spatial Graph has wrong dimension. Given {self.pc_hd.shape[1]}, but expected 4."
                    if self.tardis_logo:
                        if not assert_b:
                            TardisError(
                                id_="11",
                                py="tardis_em/utils/predictor.py",
                                desc=msg,
                            )
                            sys.exit()
                    else:
                        assert assert_b, msg
                else:
                    self.amira_image = True
                    self.image, self.px, _, self.transformation = load_am(
                        am_file=join(self.dir, id_name)
                    )
            else:
                self.amira_image = True
                self.image, self.px = load_image(join(self.dir, id_name))
                self.transformation = [0, 0, 0]
        else:
            self.image = id_name
            self.px = self.correct_px
            self.transformation = [0, 0, 0]

        self.log_prediction.append(
            f"Image pixel size: {self.px}A"
            if self.correct_px is None
            else f"Image pixel size: {self.correct_px}A"
        )
        # Normalize image histogram
        msg = f"Error while loading image {id_name}. Image loaded correctly, but output format "
        if self.amira_image and not self.binary_mask:
            self.image = self.normalize(self.mean_std(self.image)).astype(np.float32)

            # Sanity check image histogram
            if (
                not self.image.min() >= -1 or not self.image.max() <= 1
            ):  # Image not between in -1 and 1
                if self.image.min() >= 0 and self.image.max() <= 1:
                    self.image = (self.image - 0.5) * 2  # shift to -1 - 1
                elif self.image.min() >= 0 and self.image.max() <= 255:
                    self.image = self.image / 255  # move to 0 - 1
                    self.image = (self.image - 0.5) * 2  # shift to -1 - 1

            assert_b = self.image.dtype == np.float32
            if self.tardis_logo:
                if not assert_b:
                    TardisError(
                        id_="11",
                        py="tardis_em/utils/predictor.py",
                        desc=msg + f" {self.image.dtype} is not float32!",
                    )
                    sys.exit()
            else:
                assert assert_b, msg

            # Calculate parameters for image pixel size with optional scaling
            if self.correct_px is not None:
                self.px = self.correct_px

            if self.normalize_px is not None:
                self.scale_factor = self.px / self.normalize_px
            else:
                self.scale_factor = 1.0

            self.org_shape = self.image.shape
            self.scale_shape = np.multiply(self.org_shape, self.scale_factor).astype(
                np.int16
            )
            self.scale_shape = [int(i) for i in self.scale_shape]
        elif self.amira_image and self.binary_mask:
            # Check image structure
            self.image = np.where(self.image > 0, 1, 0).astype(np.int8)

            assert_b = self.image.dtype == np.int8 or self.image.dtype == np.uint8
            if self.tardis_logo:
                if not assert_b:
                    TardisError(
                        id_="11",
                        py="tardis_em/utils/predictor.py",
                        desc=msg + f" {self.image.dtype} is not int8!",
                    )
                    sys.exit()
            else:
                assert assert_b, msg

        if self.image.ndim == 2 or self.predict == "Microtubule_tirf":
            self.expect_2d = True
        else:
            self.expect_2d = False

        self.log_prediction.append(
            f"Image scaled to: {self.normalize_px}A and shape {self.scale_shape}",
        )

    def predict_cnn(self, id_i: int, id_name: str, dataloader):
        """
        Predict images using a Convolutional Neural Network (CNN) with options for image rotation
        and progress tracking integrated with the Tardis progress bar interface.

        This method iterates over a dataloader to retrieve images, predicts their output using
        a CNN model (optionally with four 90° rotations), and writes the output to `.tif` files.
        The method supports progress tracking with Tardis interface updates
        and dynamically optimizes the progress bar refresh rate based on initial iteration timing.

        :param id_i: Integer representing the ID of the image being processed.
        :type id_i: int
        :param id_name: The name of the image being processed.
        :type id_name: str
        :param dataloader: An iterable object that provides access to image data and corresponding names.
        :return: None
        """
        iter_time = 1
        if self.rotate:
            pred_title = f"CNN prediction with four 90 degree rotations with {self.convolution_nn}"
        else:
            pred_title = f"CNN prediction with {self.convolution_nn}"
        start, end = 0, 0

        for j in range(len(dataloader)):
            if j % iter_time == 0 and self.tardis_logo:
                eta_time = (
                    str(round(((end - start) * (len(dataloader) - j - 1)) / 60, 1))
                    + "min"
                )

                # Tardis progress bar update
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Found {len(self.predict_list)} images to predict! [{self.eta_predict} ETA]",
                    text_2=f"Device: {self.device}",
                    text_3=f"Image {id_i + 1}/{len(self.predict_list)} [{eta_time} ETA]: {id_name}",
                    text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                    text_5=pred_title,
                    text_7="Current Task: CNN prediction...",
                    text_8=print_progress_bar(j, len(dataloader)),
                )

            # Pick image['s]
            input_, name = dataloader.__getitem__(j)

            if j == 0:
                start = time.time()

                # Predict
                input_ = self.cnn.predict(input_[None, :], rotate=self.rotate)

                # Scale progress bar refresh to 10s
                end = time.time()

                iter_time = 10 // (end - start)
                if iter_time <= 1:
                    iter_time = 1
            else:
                # Predict
                input_ = self.cnn.predict(input_[None, :], rotate=self.rotate)
            tif.imwrite(join(self.output, f"{name}.tif"), input_)

    def predict_cnn_napari(self, input_t: torch.Tensor, name: str):
        """
        Predicts an output using the CNN model on the provided input tensor, saves the
        result in TIFF format, and returns the output tensor.

        This function performs a prediction using the Convolutional Neural Network
        (CNN) model on the given input tensor. The result is saved as a TIFF file
        using the provided file name in the specified output directory.

        :param input_t:
            Input tensor on which prediction needs to be performed, should follow
            the required input format for the CNN model.
        :param name:
            Name of the output file to save the predicted result.

        :return:
            The output tensor resulting from the CNN prediction, after processing
            with the input tensor.
        """
        input_t = self.cnn.predict(input_t[None, :], rotate=self.rotate)
        tif.imwrite(join(self.output, f"{name}.tif"), input_t)

        return input_t

    def postprocess_CNN(self, id_name: str):
        """
        Post-processes the CNN prediction by stitching predicted image patches,
        restoring the original pixel size, applying a threshold, and optionally
        saving the results in the specified format. This function also performs
        clean-up of temporary directories after processing.

        :param id_name: Identifier of the input data used to track and log the
            processed output.
        :type id_name: str
        :return: None
        """
        # Stitch predicted image patches
        if self.expect_2d:
            self.image = self.image_stitcher(
                image_dir=self.output, mask=False, dtype=np.float32
            )
            if self.image.ndim == 3:
                self.image = self.image[:, : self.scale_shape[0], : self.scale_shape[1]]
            else:
                self.image = self.image[: self.scale_shape[0], : self.scale_shape[1]]
        else:
            self.image = self.image_stitcher(
                image_dir=self.output, mask=False, dtype=np.float32
            )[: self.scale_shape[0], : self.scale_shape[1], : self.scale_shape[2]]

        # Restored original image pixel size
        self.image = torch.sigmoid(torch.from_numpy(self.image)).cpu().detach().numpy()
        self.image, _ = scale_image(image=self.image, scale=self.org_shape)

        # Threshold CNN prediction
        if self.cnn_threshold == "auto":
            self.image = adaptive_threshold(self.image)
        else:
            self.cnn_threshold = float(self.cnn_threshold)

            if self.cnn_threshold != 0:
                self.image = np.where(self.image >= self.cnn_threshold, 1, 0).astype(
                    np.uint8
                )
            else:
                if not self.output_format.startswith("return"):
                    self.log_prediction.append(
                        f"Semantic Prediction: {id_name[:-self.in_format]}"
                        f" | Number of pixels: {np.sum(self.image)}"
                    )
                    with open(join(self.am_output, "prediction_log.txt"), "w") as f:
                        f.write(" \n".join(self.log_prediction))

                    # ToDo output format specified by user aka MRC/REC/AM/TIFF
                    tif.imwrite(
                        join(self.am_output, f"{id_name[:-self.in_format]}_CNN.tif"),
                        self.image,
                    )
                    self.image = None

        """Clean-up temp dir"""
        clean_up(dir_s=self.dir)

    def preprocess_DIST(self, id_name: str):
        """
        Preprocesses a given dataset identifier (id_name) to produce and manipulate
        high-density and low-density point clouds, typically used for structural or
        image data analysis. Depending on the prediction type and the presence of
        an Amira image, this function either post-processes predicted image patches
        to construct point clouds using provided processing utilities or applies
        optimization methods like voxel down-sampling for refining existing point
        clouds.

        :param id_name: The unique dataset identifier used in debugging and processing.
        :type id_name: str
        :return: None
        """
        if self.amira_image:
            # Post-process predicted image patches
            if self.predict in ["Actin", "Microtubule", "General_filament"]:
                self.pc_hd, self.pc_ld = self.post_processes.build_point_cloud(
                    image=self.image, down_sampling=5
                )
            elif self.predict in ["Membrane2D", "Microtubule_tirf"]:
                self.pc_hd, self.pc_ld = self.post_processes.build_point_cloud(
                    image=self.image, down_sampling=5, as_2d=True
                )
            else:
                self.pc_hd, self.pc_ld = self.post_processes.build_point_cloud(
                    image=self.image, down_sampling=5, as_2d=True
                )

            self.image = None
            self._debug(id_name=id_name, debug_id="pc")
        else:
            self.pc_hd = resample_filament(self.pc_hd, self.px)
            down_sample = VoxelDownSampling(voxel=5, labels=False, KNN=True)
            self.pc_ld = down_sample(coord=self.pc_hd[:, 1:])

    def predict_DIST(self, id_i: int, id_name: str):
        """
        Predicts DIST graphs for the given coordinates using the provided DIST prediction
        model. The method processes coordinate data in chunks and updates the progress bar
        if visual feedback is enabled. The progress bar reflects the current task,
        the percentage of completion, and relevant details of the segmentation process.
        The function ensures predictive modeling for the total images with a
        controlled iteration mechanism.

        :param id_i: An integer representing the identifier of the image to be processed.
        :param id_name: A string denoting the name of the image corresponding to the ID.
        :return: A list of predicted graph representations for each coordinate dataset.
        """
        iter_time = int(round(len(self.coords_df) / 10))

        if iter_time == 0:
            iter_time = 1
        if iter_time >= len(self.coords_df):
            iter_time = 10

        pc = self.pc_ld.shape
        graphs = []
        for id_dist, coord in enumerate(self.coords_df):
            if id_dist % iter_time == 0 and self.tardis_logo:
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Found {len(self.predict_list)} images to predict! [{self.eta_predict} ETA]",
                    text_2=f"Device: {self.device}",
                    text_3=f"Image {id_i + 1}/{len(self.predict_list)}: {id_name}",
                    text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                    text_5=f"Point Cloud: {pc[0]} Nodes; NaN Segments",
                    text_7="Current Task: DIST prediction...",
                    text_8=print_progress_bar(id_dist, len(self.coords_df)),
                )

            graph = self.dist.predict(x=coord[None, :])
            graphs.append(graph)

        # Tardis progress bar update
        if self.tardis_logo:
            self.tardis_progress(
                title=self.title,
                text_1=f"Found {len(self.predict_list)} images to predict! [{self.eta_predict} ETA]",
                text_2=f"Device: {self.device}",
                text_3=f"Image {id_i + 1}/{len(self.predict_list)}: {id_name}",
                text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                text_5=f"Point Cloud: {self.pc_ld.shape[0]}; NaN Segments",
                text_7=f"Current Task: {self.predict} segmentation...",
            )

        return graphs

    def postprocess_DIST(self, id_i, i):
        """
        Processes and postprocesses data based on given inputs.

        This function adjusts the pixel data, logs information based
        on specific prediction types, and handles the transformation
        of graphs to segments. Additionally, updates the Tardis
        progress bar to provide task-specific updates.

        :param id_i: Identification number for the current image
            being processed.
        :type id_i: int

        :param i: Index of the current image.
        :type i: int

        :return: None, modifies instance attributes based
            on the processing steps.
        :rtype: None
        """
        self.pc_ld = (
            self.pc_ld * self.px
            if self.correct_px is None
            else self.pc_ld * self.correct_px
        )
        self.pc_ld = self.pc_ld + self.transformation

        if self.predict in [
            "Actin",
            "Microtubule",
            "General_filament",
            "Microtubule_tirf",
        ]:
            self.log_tardis(id_i, i, log_id=6.1)
        else:
            self.log_tardis(id_i, i, log_id=6.2)

        if self.predict in [
            "Actin",
            "Microtubule",
            "Membrane2D",
            "General_filament",
            "Microtubule_tirf",
        ]:
            sort = True
            prune = 5
        else:
            sort = False
            prune = 15

        try:
            self.segments = self.GraphToSegment.patch_to_segment(
                graph=self.graphs,
                coord=self.pc_ld,
                idx=self.output_idx,
                sort=sort,
                prune=prune,
            )
            self.segments = sort_by_length(self.segments)
        except:
            self.segments = None

        # Tardis progress bar update
        if self.tardis_logo:
            no_segments = ""
            if self.segments is None:
                no_segments = f"Point Cloud: {self.pc_ld.shape[0]}; None Segments"
            else:
                no_ = np.max(np.unique(self.segments[:, 0])).item(0)
                no_segments = f"Point Cloud: {self.pc_ld.shape[0]}; {no_} Segments"

            self.tardis_progress(
                title=self.title,
                text_1=f"Found {len(self.predict_list)} images to predict! [{self.eta_predict} ETA]",
                text_2=f"Device: {self.device}",
                text_3=f"Image {id_i + 1}/{len(self.predict_list)}: {i}",
                text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                text_5=no_segments,
                text_7=f"Current Task: {self.predict} segmented...",
            )

    def get_file_list(self):
        """
        Retrieves and processes a list of files to be used for prediction based on the directory
        or input provided. Filters files according to specified formats, handles input as either
        single directories or lists/tuples, and logs the processed files. Additionally, performs
        setup tasks for the prediction workflow, including generating paths for output directories
        and checking prediction readiness.

        :param self: The instance of the object that contains attributes such as directory paths,
                     filtering formats, continuation settings, and progress handlers for processing.
        :type self: Object containing attributes: `dir`, `available_format`, `omit_format`,
                    `continue_`, `tardis_logo`, `tardis_progress`, `output`, `am_output`,
                    `predict_list`, `device`, `title`.

        :raises AssertionError: Raised when no recognizable files exist in the provided directory
                                structure, based on specified formats, and appropriate progress
                                handling or error logging is not enabled.
        :raises Exception: Additional exceptions may occur if environmental setup or file reading
                           operations fail, depending on external utilities used and provided paths.

        :return: None
        """
        # Pickup files for the prediction
        if not isinstance(self.dir, str):
            if isinstance(self.dir, tuple) or isinstance(self.dir, list):
                self.predict_list = self.dir
            else:
                self.predict_list = [self.dir]
        else:
            if isdir(self.dir):
                self.predict_list = [
                    f
                    for f in listdir(self.dir)
                    if f.endswith(self.available_format)
                    and not f.endswith(self.omit_format)
                ]
            else:
                self.predict_list = [
                    f
                    for f in [self.dir]
                    if f.endswith(self.available_format)
                    and not f.endswith(self.omit_format)
                ]
        out_ = [
            i
            for i in self.dir.split("/")
            if not i.endswith((".mrc", ".rec", ".map", ".tif", ".tiff", ".am"))
        ]
        self.dir = join("/".join(out_))

        # Update Dir paths
        self.output = join(self.dir, "temp", "Predictions")
        self.am_output = join(self.dir, "Predictions")

        if self.continue_b:
            if isfile(join(self.am_output, "prediction_log.txt")):
                logs = np.genfromtxt(
                    join(self.am_output, "prediction_log.txt"), delimiter=",", dtype=str
                )
                logs = logs[-1].split(" ")[2]
                logs = [
                    i for i, x in enumerate(self.predict_list) if x.startswith(logs)
                ]

            if len(logs) > 0:
                logs = logs[0]
                self.predict_list = self.predict_list[logs:]

        # Check if there is anything to predict in the user-indicated folder
        msg = f"Given {self.dir} does not contain any recognizable file!"
        assert_b = len(self.predict_list) == 0

        if self.tardis_logo and self.tardis_progress is not None:
            if assert_b:
                TardisError(
                    id_="12",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
            else:
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Found {len(self.predict_list)} images to predict! [{self.eta_predict} ETA]",
                    text_2=f"Device: {self.device}",
                    text_7="Current Task: Setting-up environment...",
                )
        else:
            assert not assert_b, msg

    def log_tardis(self, id_i: int, i: Union[str, int, np.ndarray], log_id: float):
        """
        Logs various states and processing stages of the TARDIS application based on the
        provided `log_id` and input data. Depending on the log ID and input type, it generates
        log messages showcasing the progress of various computational tasks and updates a
        progress bar accordingly.

        :param id_i: Identifier for the current image being processed in the list
            of input images.
        :type id_i: int
        :param i: Input data for logging, representing either a string description
            or a numpy array. If a numpy array is passed, it is converted into a
            string representation.
        :type i: Union[str, numpy.ndarray]
        :param log_id: Numeric identifier specifying the current task or processing
            stage. Determines the type of logging information generated, and can
            optionally include different subtasks.
        :type log_id: float
        :return: None
        """
        if not isdir(self.am_output):
            mkdir(self.am_output)

        with open(join(self.am_output, "prediction_log.txt"), "w") as f:
            f.write(" \n".join(self.log_prediction))

        if isinstance(i, np.ndarray):
            i = "Numpy array"

        if not self.tardis_logo:
            return

        # Common text for all configurations
        common_text = {
            "text_1": f"Found {len(self.predict_list)} images to predict! [{self.eta_predict} ETA]",
            "text_2": f"Device: {self.device}",
            "text_3": f"Image {id_i + 1}/{len(self.predict_list)}: {i}",
        }

        if log_id == 7:
            if self.segments is not None:
                no_segments = (
                    np.max(self.segments[:, 0]) if len(self.segments) > 0 else 0
                )
            else:
                no_segments = "None"
        else:
            no_segments = "None"

        # Define text configurations for each log_id
        pc_ld_shape = self.pc_ld.shape[0]

        text_configurations = {
            0: {
                **common_text,
                "text_4": "Org. Pixel size: Nan A",
                "text_7": "Current Task: Preprocessing for CNN...",
            },
            1: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_7": f"Current Task: Sub-dividing images for {self.patch_size} size",
            },
            2: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_7": "Current Task: Stitching...",
            },
            3: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": "Point Cloud: In processing...",
                "text_7": "Current Task: Image Postprocessing...",
            },
            4: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; NaN Segments",
                "text_7": "Current Task: Preparing for instance segmentation...",
            },
            5: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; NaN Segments",
                "text_7": "Current Task: DIST prediction...",
                "text_8": (
                    print_progress_bar(0, len(self.coords_df))
                    if self.tardis_logo
                    else ""
                ),
            },
            6.1: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; NaN Segments",
                "text_7": "Current Task: Instance Segmentation...",
                "text_8": "Filament segmentation is fitted to:",
                "text_9": f"pixel size: {self.px}; transformation: {self.transformation}",
            },
            6.2: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape}; NaN Segments",
                "text_7": "Current Task: Instance segmentation...",
            },
            7: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; {no_segments} Segments",
                "text_7": "Current Task: Segmentation finished!",
            },
        }

        # Retrieve text configuration based on log_id
        config = text_configurations.get(log_id)

        # Tardis progress bar update
        self.tardis_progress(title=self.title, **config)

    def save_semantic_mask(self, i):
        """
        Saves a semantic mask prediction in a specified format and logs the prediction
        details. Supported formats include MRC, TIF, AM, and NPY. The function also
        updates a log file with prediction details and writes the semantic mask output
        to the appropriate directory in the chosen format.

        :param i: The input file name used to derive the output file name.
        :type i: str

        :raises IOError: If there are issues writing the output files or logs.
        :raises ValueError: If the specified output format is unsupported.
        """
        self.log_prediction.append(
            f"Semantic Prediction: {i[:-self.in_format]}"
            f" | Number of pixels: {np.sum(self.image)}"
        )

        with open(join(self.am_output, "prediction_log.txt"), "w") as f:
            f.write(" \n".join(self.log_prediction))

        if self.output_format.startswith("mrc"):
            to_mrc(
                data=self.image,
                file_dir=join(self.am_output, f"{i[:-self.in_format]}_semantic.mrc"),
                pixel_size=(self.px if self.correct_px is None else self.correct_px),
                label=self.semantic_header,
            )
        elif self.output_format.startswith("tif"):
            tif.imwrite(
                join(self.am_output, f"{i[:-self.in_format]}_semantic.tif"),
                (self.image if i.endswith((".mrc", ".rec")) else self.image),
            )
        elif self.output_format.startswith("am"):
            to_am(
                data=self.image,
                file_dir=join(self.am_output, f"{i[:-self.in_format]}_semantic.am"),
                pixel_size=(self.px if self.correct_px is None else self.correct_px),
                header=self.semantic_header,
            )
        elif self.output_format.startswith("npy"):
            np.save(
                join(self.am_output, f"{i[:-self.in_format]}_semantic.npy"), self.image
            )

    def save_instance_PC(self, i, overwrite_save=False):
        """
        Save processed prediction instance data to disk in various formats. This method handles
        logging, filtering, and outputting of prediction data based on the specified output format
        or other input parameters. It supports multiple output types like CSV, MRC, TIF, AM, STL,
        and NPY for different prediction types such as "Actin", "Microtubule", "Membrane",
        and general filaments or objects. Depending on the output format, it can further refine
        data through filtering, save semantic masks, or interface with Amira for spatial graph
        comparison and exportation.

        :param i: The identifier for the instance being saved.
        :type i: str
        :param overwrite_save: A flag to denote whether an existing file should be overwritten.
                               Defaults to False.
        :type overwrite_save: bool
        :return: None
        """
        if self.segments is None:
            return

        self.log_prediction.append(
            f"Instance Prediction: {i[:-self.in_format]}; Number of segments: {np.max(self.segments[:, 0])+1}"
        )

        if self.predict in [
            "Actin",
            "Microtubule",
            "General_filament",
            "Microtubule_tirf",
        ]:
            try:
                self.segments_filter = self.filter_splines(
                    segments=self.segments,
                    px=None if self.predict == "Microtubule_tirf" else self.px,
                )
                self.segments_filter = sort_by_length(self.segments_filter)

                self.log_prediction.append(
                    f"Instance Prediction: {i[:-self.in_format]};"
                    f" Number of segments: {np.max(self.segments_filter[:, 0])+1}"
                )
            except ValueError:
                self.segments_filter = None

        with open(join(self.am_output, "prediction_log.txt"), "w") as f:
            f.write(" \n".join(self.log_prediction))

        if self.output_format.endswith("amSG"):
            if self.predict in [
                "Actin",
                "Microtubule",
                "General_filament",
                "Microtubule_tirf",
            ]:
                self.amira_file.export_amira(
                    coords=self.segments,
                    file_dir=join(
                        self.am_output, f"{i[:-self.in_format]}_SpatialGraph.am"
                    ),
                    labels=["TardisPrediction"],
                    scores=[
                        ["EdgeLength", "EdgeConfidenceScore"],
                        [
                            length_list(self.segments),
                            self.score_splines(self.segments),
                        ],
                    ],
                )

                if self.segments_filter is not None:
                    self.amira_file.export_amira(
                        coords=self.segments_filter,
                        file_dir=join(
                            self.am_output,
                            f"{i[:-self.in_format]}_SpatialGraph_filter.am",
                        ),
                        labels=["TardisPrediction"],
                        scores=[
                            ["EdgeLength", "EdgeConfidenceScore"],
                            [
                                length_list(self.segments_filter),
                                self.score_splines(self.segments_filter),
                            ],
                        ],
                    )

                if self.amira_check and self.predict == "Microtubule":
                    dir_amira_file = join(
                        self.dir_amira,
                        i[: -self.in_format] + self.amira_prefix + ".am",
                    )

                    if isfile(dir_amira_file):
                        amira_sg = ImportDataFromAmira(src_am=dir_amira_file)
                        amira_sg = amira_sg.get_segmented_points()

                        if amira_sg is not None:
                            if self.segments_filter is not None:
                                compare_sg, label_sg = self.compare_spline(
                                    amira_sg=amira_sg, tardis_sg=self.segments_filter
                                )
                            else:
                                compare_sg, label_sg = self.compare_spline(
                                    amira_sg=amira_sg, tardis_sg=self.segments
                                )

                            if self.output_format.endswith("amSG"):
                                self.amira_file.export_amira(
                                    file_dir=join(
                                        self.am_output,
                                        f"{i[:-self.in_format]}_AmiraCompare.am",
                                    ),
                                    coords=compare_sg,
                                    labels=label_sg,
                                    scores=None,
                                )
            else:
                self.amira_file_points.export_amira(
                    coords=self.segments,
                    file_dir=join(
                        self.am_output, f"{i[:-self.in_format]}_SpatialGraph.am"
                    ),
                    labels=["TardisPrediction"],
                )
                if self.segments_filter is not None:
                    self.amira_file_points.export_amira(
                        coords=self.segments_filter,
                        file_dir=join(
                            self.am_output,
                            f"{i[:-self.in_format]}_SpatialGraph_filter.am",
                        ),
                        labels=["TardisPrediction"],
                    )
        elif self.output_format.endswith("csv") or overwrite_save:
            segments = pd.DataFrame(self.segments)
            segments.to_csv(
                join(self.am_output, f"{i[:-self.in_format]}_instances.csv"),
                header=["IDs", "X [A]", "Y [A]", "Z [A]"],
                index=False,
                sep=",",
            )

            if self.predict in [
                "Actin",
                "Microtubule",
                "Microtubule_tirf",
                "Membrane2D",
                "General_filament",
            ]:
                if self.segments_filter is not None:
                    if len(self.segments_filter) > 0:
                        self.segments_filter = pd.DataFrame(self.segments_filter)
                        self.segments_filter.to_csv(
                            join(
                                self.am_output,
                                f"{i[:-self.in_format]}_instances_filter.csv",
                            ),
                            header=["IDs", "X [A]", "Y [A]", "Z [A]"],
                            index=False,
                            sep=",",
                        )
                        self.segments_filter = self.segments_filter.to_numpy()
        elif self.output_format.endswith(("mrc", "tif", "am")):
            if self.predict in [
                "Membrane",
                "Membrane2D",
                "General_object",
                "Microtubule_tirf",
            ]:
                self.mask_semantic = draw_semantic_membrane(
                    mask_size=self.org_shape,
                    coordinate=self.segments,
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    spline_size=(
                        60
                        if self.predict
                        not in [
                            "Microtubule_tirf",
                            "Microtubule",
                            "General_filament",
                            "Actin",
                        ]
                        else 25
                    ),
                )
            else:
                self.segments[:, 1:] = (
                    self.segments[:, 1:] / self.px
                    if self.correct_px is None
                    else self.correct_px
                )

                self.mask_semantic = draw_instances(
                    mask_size=self.org_shape,
                    coordinate=self.segments,
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    circle_size=125,
                )
            self._debug(id_name=i, debug_id="instance_mask")

            if self.output_format.endswith("mrc"):
                to_mrc(
                    data=self.mask_semantic,
                    file_dir=join(
                        self.am_output, f"{i[:-self.in_format]}_instance.mrc"
                    ),
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    label=self.instance_header,
                )
            elif self.output_format.endswith("tif"):
                tif.imwrite(
                    join(self.am_output, f"{i[:-self.in_format]}_instance.tif"),
                    self.mask_semantic,
                )
            elif self.output_format.endswith("am"):
                to_am(
                    data=self.mask_semantic,
                    file_dir=join(self.am_output, f"{i[:-self.in_format]}_instance.am"),
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    header=self.instance_header,
                )
        elif self.output_format.endswith("stl"):
            if self.predict == "Membrane":
                to_stl(
                    data=self.segments,
                    file_dir=join(self.am_output, f"{i[:-self.in_format]}.stl"),
                )
        elif self.output_format.endswith("npy"):
            np.save(
                join(self.am_output, f"{i[:-self.in_format]}_instance.npy"),
                self.segments,
            )

    def _debug(self, id_name: str, debug_id: str):
        """
        Executes specific debug procedures based on the provided debug identifier. Depending
        on the `debug_id` and other instance properties, it processes and writes files
        including TIFF images, NPY arrays, or other data outputs to the defined output
        directory.

        :param id_name: Name or identifier for the current operation or file being processed.
        :param debug_id: Debug identifier to specify the type of debug operation to perform.
                         Acceptable values include "cnn", "pc", "graph", "segment", and
                         "instance_mask".
        :return: None
        """
        if self.debug:
            if debug_id == "cnn":
                tif.imwrite(
                    join(self.am_output, f"{id_name[:-self.in_format]}_CNN.tif"),
                    self.image,
                )
            elif debug_id == "pc":
                np.save(
                    join(self.am_output, f"{id_name[:-self.in_format]}_raw_pc_hd.npy"),
                    self.pc_hd,
                )
                np.save(
                    join(self.am_output, f"{id_name[:-self.in_format]}_raw_pc_ld.npy"),
                    self.pc_ld,
                )
            elif debug_id == "graph":
                try:
                    np.save(
                        join(
                            self.am_output,
                            f"{id_name[:-self.in_format]}_graph_voxel.npy",
                        ),
                        self.graphs,
                    )
                except ValueError:
                    pass
            elif debug_id == "segment":
                if self.device == "cpu":
                    np.save(
                        join(
                            self.am_output,
                            f"{id_name[:-self.in_format]}_coord_voxel.npy",
                        ),
                        self.coords_df,
                    )
                    np.save(
                        join(
                            self.am_output, f"{id_name[:-self.in_format]}_idx_voxel.npy"
                        ),
                        self.output_idx,
                    )
                else:
                    try:
                        np.save(
                            join(
                                self.am_output,
                                f"{id_name[:-self.in_format]}_coord_voxel.npy",
                            ),
                            self.coords_df.cpu().detach().numpy(),
                        )
                    except AttributeError:
                        pass
                    try:
                        np.save(
                            join(
                                self.am_output,
                                f"{id_name[:-self.in_format]}_idx_voxel.npy",
                            ),
                            self.output_idx.cpu().detach().numpy(),
                        )
                    except:
                        pass
                np.save(
                    join(self.am_output, f"{id_name[:-self.in_format]}_segments.npy"),
                    self.segments,
                )
            elif debug_id == "instance_mask":
                np.save(
                    join(
                        self.am_output, f"{id_name[:-self.in_format]}_instance_mask.npy"
                    ),
                    self.mask_semantic,
                )

    def __call__(self, save_progres=False):
        """
        Executes the object as a callable, processing and predicting on a dataset of files
        using a combination of semantic segmentation and instance segmentation workflows.
        The function involves data loading, pre-processing, neural network (CNN) inference,
        post-processing, and saving results. Predictions can be returned optionally if
        specified in the output format.

        The function supports multiple input formats and provides options for binary masks,
        instance-level predictions, and debug checkpoints. Additionally, it handles errors,
        progress updates via a Tardis logger, and optional clean-up of temporary directories.
        Intermediate and final results, like semantic masks or instance segmentations
        from DIST predictions, are saved and optionally returned for further analysis.

        :param save_progres: Boolean flag indicating whether to overwrite and save the
            progress of processing for each file. Default is False.
        :type save_progres: bool

        :return: Depending on the output format, returns semantic segmentation outputs,
            instance segmentation outputs, and instance-filtered outputs in the form
            of lists, where each list entry corresponds to the prediction of an individual file.
        """
        self.get_file_list()

        semantic_output, instance_output, instance_filter_output = [], [], []
        for id_, i in enumerate(self.predict_list):
            start_predict = time.time()

            """CNN Pre-Processing"""
            if isinstance(i, str):
                # Find a file format
                self.in_format = 0
                if i.endswith((".tif", ".mrc", ".rec", ".map", ".npy")):
                    self.in_format = 4
                elif i.endswith(".tiff"):
                    self.in_format = 5
                elif i.endswith(".am"):
                    self.in_format = 3

            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=0)

            # Load data
            self.load_data(id_name=i)

            msg = (
                f"Predicted file {id_} is numpy array without pixel size metadate {self.px}."
                "Please pass correct_px argument as a correct pixel size value."
            )
            assert_b = self.px is None and not isinstance(i, str)
            if assert_b:
                if self.tardis_logo:
                    TardisError(id_="161", py="tardis_em.utils.predictor.py", desc=msg)
                    sys.exit()
                else:
                    assert not assert_b, msg

            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=1)

            """Semantic Segmentation"""
            if not self.binary_mask:
                trim_with_stride(
                    image=self.image,
                    scale=self.scale_shape,
                    trim_size_xy=self.patch_size,
                    trim_size_z=self.patch_size,
                    output=join(self.dir, "temp", "Patches"),
                    image_counter=0,
                    clean_empty=False,
                    stride=round(self.patch_size * 0.125),
                )
                self.image = None

                """CNN prediction"""
                self.predict_cnn(
                    id_i=id_,
                    id_name=i,
                    dataloader=PredictionDataset(
                        join(self.dir, "temp", "Patches", "imgs")
                    ),
                )

                """CNN Post-Processing"""
                # Tardis progress bar update
                self.log_tardis(id_, i, log_id=2)

                # Post-process
                self.postprocess_CNN(id_name=i)

                # Store optionally for return
                if self.output_format.startswith("return"):
                    semantic_output.append(self.image)

                # Check if the image is binary
                if self.image is not None and not self.output_format.startswith(
                    "return"
                ):
                    if len(pd.unique(self.image.flatten())) == 1:
                        self.image = None
                        self.log_prediction.append(
                            f"Semantic prediction: No semantic detected."
                        )
                    else:
                        end_predict = time.time()
                        half_time = round((end_predict - start_predict) / 60, 1)
                        self.log_prediction.append(
                            f"Semantic Prediction Finished in: {half_time} min"
                        )

                    with open(join(self.am_output, "prediction_log.txt"), "w") as f:
                        f.write(" \n".join(self.log_prediction))

                if self.image is None:
                    continue

                # Debug flag
                self._debug(id_name=i, debug_id="cnn")

                # Check if predicted image
                assert_b = self.image.shape == self.org_shape
                if not assert_b:
                    msg = (
                        "Last Task: Stitching/Scaling/Make correction..."
                        f"Tardis Error: Error while converting to {self.px} A "
                        f"Org. shape {self.org_shape} is not the same as "
                        f"converted shape {self.image.shape}"
                    )
                    if self.tardis_logo:
                        TardisError(
                            id_="116",
                            py="tardis_em/utils/predictor.py",
                            desc=msg,
                        )
                    else:
                        assert assert_b, msg
                    sys.exit()

                # Save predicted mask as file
                if self.image.ndim == 3:
                    _, y_dim, x_dim = self.image.shape
                else:
                    y_dim, x_dim = self.image.shape
                y_pad = max(1, min(5, int(round(y_dim * 0.01))))
                x_pad = max(1, min(5, int(round(x_dim * 0.01))))

                self.image[..., :x_pad] = 0
                self.image[..., -x_pad:] = 0
                self.image[..., :y_pad, :] = 0
                self.image[..., -y_pad:, :] = 0

                self.save_semantic_mask(i)

                # Sanity check for binary mask
                if not self.image.min() == 0 and not self.image.max() == 1:
                    continue

                # Close iter loop if not instance prediction
                if not self.predict_instance:
                    continue

            """Instance Segmentation"""
            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=3)

            self.preprocess_DIST(id_name=i)
            self.segments = np.zeros((0, 4))
            self.segments_filter = None

            # Break iter loop for instances if no point cloud is found
            if len(self.pc_ld) == 0:
                self.log_prediction.append(
                    f"Instance prediction: Not enough point from semantic mask were generated."
                )

                if self.output_format.endswith("return"):
                    instance_output.append(np.zeros((0, 4)))
                    instance_filter_output.append(np.zeros((0, 4)))
                continue

            with open(join(self.am_output, "prediction_log.txt"), "w") as f:
                f.write(" \n".join(self.log_prediction))

            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=4)

            # Build patches dataset
            (
                self.coords_df,
                _,
                self.output_idx,
                _,
            ) = self.patch_pc.patched_dataset(coord=self.pc_ld)

            # Predict point cloud
            self.log_tardis(id_, i, log_id=5)

            # DIST prediction
            self.graphs = self.predict_DIST(id_i=id_, id_name=i)
            self._debug(id_name=i, debug_id="graph")
            # Save debugging check point
            self._debug(id_name=i, debug_id="segment")

            # DIST Instance graph-cut
            self.postprocess_DIST(id_, i)

            if self.segments is None:
                self.log_prediction.append(
                    f"Instance prediction: Not segments could be predicted with point cloud generated from Semantic mask."
                )
            else:
                half_time = round((time.time() - end_predict) / 60, 1)
                self.log_prediction.append(
                    f"Instance Prediction Finished in: {half_time} min"
                )

            self.log_prediction.append("\n")
            with open(join(self.am_output, "prediction_log.txt"), "w") as f:
                f.write(" \n".join(self.log_prediction))

            self.log_tardis(id_, i, log_id=7)

            """Save as .am"""
            self.save_instance_PC(i, overwrite_save=save_progres)

            if self.segments is None:
                if self.output_format.endswith("return"):
                    instance_output.append(np.zeros((0, 4)))
                    instance_filter_output.append(np.zeros((0, 4)))
            else:
                instance_output.append(self.segments)
                if self.predict in [
                    "Actin",
                    "Microtubule",
                    "General_filament",
                    "Microtubule_tirf",
                ]:
                    instance_filter_output.append(self.segments_filter)

            """Clean-up temp dir"""
            clean_up(dir_s=self.dir)
            end_predict = time.time()

            self.eta_predict = (
                str(
                    round(
                        (
                            (end_predict - start_predict)
                            * (len(self.predict_list) - id_ - 1)
                        )
                        / 60,
                        1,
                    )
                )
                + " min"
            )

        """Optional return"""
        if self.output_format.startswith("return"):
            if self.output_format.endswith("return"):
                return semantic_output, instance_output, instance_filter_output
            return semantic_output
        else:
            return


class Predictor:
    """
    Handles model prediction workflows for neural networks, including loading pretrained
    weights, configuring model architectures dynamically, and predicting data. The
    purpose of this class is to abstract away the complexities of network setup and
    enhance user focus on utilizing pretrained networks, streamlining predictions.

    The class ensures compatibility with various deep learning frameworks, supports
    CNNs and distance-based networks dynamically, and provides inference-time adjustments like
    rotations for robustness.
    """

    def __init__(
        self,
        device: torch.device,
        network: Optional[str] = None,
        checkpoint: Optional[str] = None,
        subtype: Optional[str] = None,
        model_version: Optional[int] = None,
        img_size: Optional[int] = None,
        model_type: Optional[str] = None,
        sigma: Optional[float] = None,
        sigmoid=True,
        _2d=False,
        logo=True,
    ):
        """
        This class initializer is responsible for setting up a model in the
        TARDIS framework. It includes initializing the model's structure,
        loading weights from a checkpoint or AWS, configuring model parameters,
        and setting up essential model properties like sigmoid activation,
        coordinate embeddings, and device placement.

        :param device: Specifies the torch device (e.g., CPU or GPU)
                       on which the model will be loaded.
        :type device: torch.device
        :param network: The name of the network to be used. If None, the
                        function will automatically determine it based on
                        the model structure.
        :type network: Optional[str]
        :param checkpoint: Specifies the path to the model checkpoint file or
                           directly a preloaded weights dictionary. If not
                           provided, it attempts to find weights on AWS.
        :type checkpoint: Optional[str]
        :param subtype: Defines the network subtype to search specific weights.
                        This is typically used to distinguish between variants
                        of the same base network.
        :type subtype: Optional[str]
        :param model_version: Indicates the model version to use for locating
                              specific weights on AWS.
        :type model_version: Optional[int]
        :param img_size: Determines the image size to which the model will
                         process input. This is primarily for patch-based CNNs.
        :type img_size: Optional[int]
        :param model_type: The type or category of the model (e.g., cnn or dist).
        :type model_type: Optional[str]
        :param sigma: Used for setting the sigma value for coordinate embedding
                      if provided. It can also overwrite the value in weights
                      dictionary.
        :type sigma: Optional[float]
        :param sigmoid: A boolean flag to determine whether sigmoid activation
                        should be applied to the model's output.
        :type sigmoid: bool
        :param _2d: Indicates whether the model processes 2D data (True) or
                    3D data (False).
        :type _2d: bool
        :param logo: A flag that controls whether to enforce specific checks
                     like model network and weights validation on initialization.
        :type logo: bool

        :raises TardisError: Custom TARDIS-related error when neither network
                             nor checkpoint is provided in certain scenarios.
        :raises AssertionError: Raised when logo is False and both network and
                                checkpoint parameters are not defined.
        """
        self.logo = logo

        self.device = device
        self.img_size = img_size

        msg = "Missing network weights or network name!"
        if self.logo:
            if checkpoint is None and network is None:
                TardisError(
                    "139",
                    "tardis_em/utils/predictor.py",
                    msg,
                )
        else:
            assert checkpoint is not None and network is not None, msg

        if checkpoint is None:
            print(
                f"Searching for weight file for {network}_{subtype} for {model_type}..."
            )

            weights = torch.load(
                get_weights_aws(network, subtype, model_type, model_version),
                map_location="cpu",
                weights_only=False,
            )
        elif isinstance(checkpoint, dict):
            print("Loading weight dictionary...")
            weights = checkpoint
        else:
            print("Loading weight model...")
            weights = torch.load(checkpoint, map_location="cpu", weights_only=False)

        # Load model
        if isinstance(weights, dict):  # Deprecated support for dictionary
            # Allow overwriting sigma
            if sigma is not None:
                weights["model_struct_dict"]["coord_embed_sigma"] = sigma

            if "dist_type" in weights["model_struct_dict"]:
                from tardis_em.dist_pytorch.utils.utils import check_model_dict
            else:
                from tardis_em.cnn.utils.utils import check_model_dict

            model_structure = check_model_dict(weights["model_struct_dict"])

            if network is not None:
                self.network = network
            else:
                if "dist_type" in model_structure:
                    self.network = "dist"
                else:
                    self.network = "cnn"

            self._2d = _2d
            self.model = self._build_model_from_checkpoint(
                structure=model_structure, sigmoid=sigmoid
            )
            self.model.load_state_dict(weights["model_state_dict"])
        else:  # Load onnx or another model
            self.network = network
            self._2d = _2d
            self.model = weights

            if not network == "dist":
                self.model.update_patch_size(self.img_size, sigmoid)
        self.model.to(self.device)
        self.model.eval()

        del weights  # Cleanup weight file from memory

    def _build_model_from_checkpoint(self, structure: dict, sigmoid=True):
        """
        Builds a model from the given checkpoint structure.

        This method constructs a machine learning model based on a provided
        configuration dictionary. The configuration can specify different
        types of networks such as `dist_type` or `cnn_type`. In case the
        structure does not match these options, it will return `None`.
        For `dist_type`, it creates a distribution-based network, while for
        `cnn_type`, it constructs a CNN-based network.

        :param structure: A dictionary containing the model configuration
                          blueprint, including the type of network
                          (`dist_type` or `cnn_type`) and other necessary
                          parameters.
        :type structure: dict
        :param sigmoid: A boolean indicating whether the model's final
                        output should apply a sigmoid function or not.
                        Defaults to ``True``.
        :type sigmoid: bool
        :return: A constructed model instance based on the specified
                 structure and type, or ``None`` if no valid network type
                 is provided.
        :rtype: object | None
        """
        if "dist_type" in structure:
            model = build_dist_network(
                network_type=structure["dist_type"],
                structure=structure,
                prediction=sigmoid,
            )
        elif "cnn_type" in structure:
            model = build_cnn_network(
                network_type=structure["cnn_type"],
                structure=structure,
                img_size=self.img_size,
                prediction=sigmoid,
            )
        else:
            model = None

        return model

    def predict(
        self, x: torch.Tensor, y: Optional[torch.Tensor] = None, rotate=False
    ) -> np.ndarray:
        """
        Predicts an output based on given input data using a trained model. The method supports
        various modes of operation, including computing outputs for specific network types or
        applying rotations to the input for models with two-dimensional or three-dimensional
        spatial components. This function can handle data provided as PyTorch tensors or convert
        NumPy arrays into tensors internally. Outputs are generated either in a transformed or
        direct form according to the network type and additional parameters provided.

        :param x: Input tensor containing the primary data for prediction. Expected to have
                  shapes conforming to the model's requirements.
        :type x: torch.Tensor
        :param y: Optional secondary input tensor containing additional data or node features.
                  Defaults to None. Expected to match the compatible input feature dimensions
                  of the model if provided.
        :type y: Optional[torch.Tensor]
        :param rotate: Boolean flag indicating whether rotations should be applied to the input
                       tensor to generate averaged transformed outputs. Defaults to False.
        :type rotate: bool
        :return: The processed output from the model, structured as a NumPy array. Its dimensions
                 and content correspond to the defined task of the provided trained model. Output
                 is adjusted according to whether rotations were applied or if specific
                 computations are required by the network type.
        :rtype: np.ndarray
        """
        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        if y is not None and isinstance(y, np.ndarray):
            y = torch.Tensor(y)

        with torch.no_grad():
            dim_ = x.shape[-1]
            x = x.to(self.device)

            if self.network == "dist":
                if y is None:
                    out = self.model(coords=x, node_features=None)
                else:
                    out = self.model(coords=x, node_features=y.to(self.device))

                out = out.cpu().detach().numpy()[0, 0, :]
                g_len = out.shape[0]
                g_range = range(g_len)

                # Overwrite diagonal with 1
                out[g_range, g_range] = np.eye(g_len, g_len)[g_range, g_range]
                return out
            else:
                if rotate:
                    if self._2d:
                        out = np.zeros((dim_, dim_), dtype=np.float32)
                        for k in range(4):
                            x_ = torch.rot90(x, k=k, dims=(2, 3))
                            x_ = self.model(x_) / 4
                            x_ = x_.cpu().detach().numpy()[0, 0, :]

                            out += np.rot90(x_, k=-k, axes=(0, 1))
                    else:
                        out = np.zeros((dim_, dim_, dim_), dtype=np.float32)
                        for k in range(4):
                            x_ = torch.rot90(x, k=k, dims=(3, 4))
                            x_ = self.model(x_) / 4
                            x_ = x_.cpu().detach().numpy()[0, 0, :]

                            out += np.rot90(x_, k=-k, axes=(1, 2))
                else:
                    out = self.model(x).cpu().detach().numpy()[0, 0, :]

                return out



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
import torch
import torch.nn as nn


class SparsTriangularUpdate(nn.Module):
    """
    Implements a neural network module for sparsely updating tensors using a triangular update rule.

    The SparsTriangularUpdate class provides a mechanism to apply sparse triangular updates on
    coordinate tensors, enabling efficient computations for specific data structures or patterns.
    The module includes normalization layers, linear transformations, and gating mechanisms
    to modulate and compute updates. It supports customizable parameters for dimensions
    and the axis of operations, as well as a k-nearest-neighbor (knn) connectivity.
    """

    def __init__(self, input_dim: int, channel_dim=128, axis=1, knn=8):
        """
        Class instantiates and configures layers and parameters for a neural network
        module. The layers include LayerNorm and Linear transformations applied to
        inputs based on dimensions specified during initialization. The class
        also ensures proper initialization of parameters to achieve stability.

        :param input_dim: Input feature dimension for the given data.
        :type input_dim: int
        :param channel_dim: Number of output channels for the linear layers. Defaults to 128.
        :type channel_dim: int, optional
        :param axis: Axis along which normalization will be performed. Defaults to 1.
        :type axis: int, optional
        :param knn: Number of nearest neighbors to consider for operations (specific
                    use case dependent). Defaults to 8.
        :type knn: int, optional
        """
        super().__init__()
        self.input_dim = input_dim
        self.channel_dim = channel_dim
        self.axis = axis
        self.init_scaling = 1 / 1.4142135623730951
        self.knn = knn

        # Define the transformations to be applied
        self.norm_input = nn.LayerNorm(input_dim)

        self.linear_a = nn.Linear(input_dim, channel_dim)
        self.gate_a = nn.Linear(input_dim, channel_dim)

        self.linear_b = nn.Linear(input_dim, channel_dim)
        self.gate_b = nn.Linear(input_dim, channel_dim)

        self.norm_o = nn.LayerNorm(channel_dim)
        self.gate_o = nn.Linear(input_dim, input_dim)
        self.linear_o = nn.Linear(channel_dim, input_dim)

        self._reset_parameters()

    def _reset_parameters(self):
        """
        Initial parameter and bias scaling.
        """
        nn.init.xavier_uniform_(self.linear_a.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_a.bias, 0.0)

        nn.init.xavier_uniform_(self.linear_b.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_b.bias, 0.0)

        nn.init.constant_(self.linear_o.weight, 0.0)
        nn.init.constant_(self.linear_o.bias, 0.0)

    def forward(self, x: torch.tensor, indices: list) -> torch.tensor:
        """
        Processes input tensor `x` using gated linear transformations and triangular
        multiplication updates with specified row-wise or column-wise operations,
        returning the updated tensor as a result.

        This function applies the following transformations:
        1. Normalizes the input tensor `x`.
        2. Computes intermediate transformations `a` and `b` using sigmoid activations
           and a combination of gate layers (`gate_a`, `gate_b`) along with linear
           layers (`linear_a`, `linear_b`).
        3. Depending on the specified axis, computes either a row-wise or column-wise
           multiplication update using the indices provided.
        4. Applies a final sigmoid activation combined with gated output
           transformation using `gate_o` and `linear_o`.

        :param x: The input tensor. Shape `[N, Ch]` representing
            batch size (`N`) by the number of channels (`Ch`).
        :param indices: List of index tensors used to specify the
            range for row-wise or column-wise updates. For row updates,
            indices[0] is used, and for column updates, indices[1] is used.
        :return: Updated tensor after applying gated transformations
            and triangular multiplication updates. Shape `[N, Ch]`.
        """
        x = self.norm_input(x)  # Batch x Length x Channels [N, Ch]

        # Compute intermediate transformations
        a = torch.sigmoid(self.gate_a(x)) * self.linear_a(x)
        b = torch.sigmoid(self.gate_b(x)) * self.linear_b(x)

        # Apply triangular multiplication update

        if self.axis == 1:  # Row-wise update
            k = torch.einsum("ik,ijk->ik", a, b[indices[0]])
        else:  # Col-wise update
            k = torch.einsum("ik,ijk->ik", a, b[indices[1]])

        return torch.sigmoid(self.gate_o(x)) * self.linear_o(self.norm_o(k))



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
import time
from os import listdir
from os.path import expanduser, join
from typing import Optional

import click
import torch

from tardis_em.benchmarks.predictor import CnnBenchmark, DISTBenchmark
from tardis_em.utils.aws import get_benchmark_aws, put_benchmark_aws
from tardis_em.utils.device import get_device
from tardis_em.utils.errors import TardisError
from tardis_em.utils.logo import TardisLogo
from tardis_em.utils.metrics import compare_dict_metrics
from tardis_em.utils.predictor import Predictor
from tardis_em._version import version


@click.command()
@click.option(
    "-dir",
    "--local_directory",
    default=None,
    type=str,
    help="Optionally, specified benchmark directory if different then"
    "SMLC local server.",
    show_default=True,
)
@click.option(
    "-ds",
    "--data_set",
    type=str,
    help="Data set name used for testing, should be the same as dataset"
    "storage folder name for given dataset.",
    show_default=True,
)
@click.option(
    "-ch",
    "--model_checkpoint",
    type=str,
    help="Directory for model pre-trained weight and structure dictionary.",
    show_default=True,
)
@click.option(
    "-th",
    "--nn_threshold",
    default=0.5,
    type=float,
    help="Threshold use for NN prediction.",
    show_default=True,
)
@click.option(
    "-ps",
    "--patch_size",
    default=None,
    type=int,
    help="Optional: Size of image size used for prediction.",
    show_default=True,
)
@click.option(
    "-sg",
    "--sigma",
    default=None,
    type=float,
    help="Optional: Sigma value for distance embedding.",
    show_default=True,
)
@click.option(
    "-pv",
    "--points_in_patch",
    default=1000,
    type=int,
    help="Optional: Number of point per voxel.",
    show_default=True,
)
@click.option(
    "-sv",
    "--save",
    default=True,
    type=bool,
    help="Flag to switch of saving model weights and metric on S3.",
    show_default=True,
)
@click.option(
    "-d",
    "--device",
    default="0",
    type=str,
    help="Define which device use for training: "
    "0-9: Use positive numeric value to use specific GPU ID "
    "-1: Usa to use CPU "
    "mps: Use to use Apple Silicon",
    show_default=True,
)
@click.version_option(version=version)
def main(
    local_directory: Optional[str],
    data_set: str,
    model_checkpoint: str,
    nn_threshold: float,
    patch_size: Optional[int],
    sigma: Optional[float],
    points_in_patch: Optional[int],
    save: bool,
    device: str,
):
    """
    Standard benchmark for DIST on medical and standard point clouds
    """
    """Global setting"""
    tardis_progress = TardisLogo()
    title = "TARDIS - NN Benchmark"
    tardis_progress(title=title)

    # Specified local directory for benchmark dataset
    if local_directory is None:  # Default NYSBC-SMLC internal server
        DIR_ = join(expanduser("~") + "/../../data/rkiewisz/Benchmarks")
    else:
        DIR_ = local_directory

    """Get model for benchmark"""
    model = torch.load(
        model_checkpoint, map_location=get_device(device), weights_only=False
    )

    """Best model list from S3"""
    rgb = False
    if data_set.endswith("rgb"):
        data_set = data_set[:-4]
        rgb = True

    if [True for x in model["model_struct_dict"] if x.startswith("cnn")]:
        network = "cnn"
        DIR_NN = join(DIR_, "Best_model_CNN", data_set)
        DIR_EVAL = join(DIR_, "Eval_CNN", data_set)
    else:
        network = "dist"
        DIR_NN = join(DIR_, "Best_model_DIST", data_set)
        DIR_EVAL = join(DIR_, "Eval_DIST", data_set)

    if data_set not in listdir(join(DIR_, "Eval_DIST")):
        TardisError(
            id_="",
            py="tardis_em/benchmarks/benchmarks.py",
            desc=f"Given data set {data_set} is not supporter! "
            f'Expected one of {listdir(join(DIR_, "Eval_DIST"))}',
        )

    if rgb:
        data_set = f"{data_set}_rgb"

    BEST_SCORE = get_benchmark_aws()

    m_name = model["model_struct_dict"]

    predictor = Predictor(
        checkpoint=model, img_size=patch_size, sigma=sigma, device=get_device(device)
    )

    """Build DataLoader"""
    if network == "cnn":
        m_name = f'{m_name["cnn_type"]}_{m_name["conv_scaler"]}'

        predictor_bch = CnnBenchmark(
            model=predictor,
            dataset=data_set,
            dir_s=DIR_EVAL,
            threshold=nn_threshold,
            patch_size=patch_size,
        )
    else:
        m_name = f'dist_{m_name["dist_type"]}_{m_name["structure"]}'

        predictor_bch = DISTBenchmark(
            model=predictor,
            dataset=data_set,
            dir_s=DIR_EVAL,
            threshold=nn_threshold,
            points_in_patch=points_in_patch,
        )
    nbm = predictor_bch()

    """Compared with best models"""
    new_is_best = False
    model_best_time = "None"
    cbm = None
    model_id = 0

    # Pick last best model
    try:
        model_best = BEST_SCORE[m_name][data_set]
        model_id = len(model_best)
        model_best = model_best[-1]  # Pick last/best model from the list
        model_best_time = model_best[0]  # Pick benchmark time
        cbm = model_best[3]  # Pick the best metric dictionary
    except KeyError:
        pass

    """Benchmark Summary"""
    metric_keys = (
        ["IoU", "AUC", "mCov", "mWCov"]
        if network != "cnn"
        else ["F1", "AUC", "IoU", "AP"]
    )

    if model_best_time != "None":
        # Take mean of all metrics and check which one performs better
        new_is_best = compare_dict_metrics(last_best_dict=cbm, new_dict=nbm)

        best_metric = "; ".join([f"{key}: {cbm.get(key, '')}" for key in metric_keys])
    else:
        best_metric = "There is no existing model to compare with."

    new_metric = "; ".join([f"{key}: {nbm.get(key, '')}" for key in metric_keys])

    tardis_progress(
        title="TARDIS - NN Benchmark - Results",
        text_1=f"New model is better: {new_is_best}",
        text_3=f"Benchmark results for model: {m_name}: ",
        text_4=f"Model dataset benchmarked on: {data_set}",
        text_6=f"Last best model from [{model_best_time}]:",
        text_7=best_metric,
        text_9=f"Current benchmark model from [{time.asctime()}]:",
        text_10=new_metric,
    )

    """Sent updated json and model to S3"""
    if new_is_best:
        id_save = f"{m_name}_{data_set}_{model_id}"
    else:
        id_save = f"{m_name}_{data_set}_0"
    link = (
        "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/benchmark/models/"
        f"{id_save}"
    )

    if save:
        if m_name in BEST_SCORE:
            if data_set in BEST_SCORE[m_name]:
                BEST_SCORE[m_name][data_set].append(
                    [time.asctime(), link, round(sum(nbm.values()) / len(nbm), 2), nbm]
                )
            else:
                BEST_SCORE[m_name][data_set] = [
                    [time.asctime(), link, round(sum(nbm.values()) / len(nbm), 2), nbm]
                ]
        else:
            BEST_SCORE[m_name] = {f"{data_set}": []}
            BEST_SCORE[m_name][data_set].append(
                [time.asctime(), link, round(sum(nbm.values()) / len(nbm), 2), nbm]
            )

        """Upload model and json"""
        if model_best_time == "None":
            put_benchmark_aws(data=BEST_SCORE, network=id_save, model=model_checkpoint)
            torch.save(model_checkpoint, join(DIR_NN, id_save))
        elif new_is_best:
            put_benchmark_aws(data=BEST_SCORE, network=id_save, model=model_checkpoint)
            torch.save(model_checkpoint, join(DIR_NN, id_save))


if __name__ == "__main__":
    main()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

import shutil
from os.path import join

import numpy as np

from tardis_em.dist_pytorch.datasets.augmentation import preprocess_data
from tardis_em.dist_pytorch.datasets.dataloader import (
    FilamentDataset,
    PartnetDataset,
    ScannetColorDataset,
    ScannetDataset,
)


class TestDataLoader:
    dir_ = join("tests", "test_data", "data_loader")
    coord_dir = join(dir_, "filament_mt", "train", "masks", "am3D.CorrelationLines.am")
    img_dir = join("tests", "test_data", "data_loader")

    def test_preprocess_3d_img(self):
        coord, img = preprocess_data(
            coord=self.coord_dir,
            image=join(self.img_dir, "filament_mt", "train", "imgs", "am3D.am"),
            size=64,
        )
        assert coord.ndim == 2
        assert coord.shape == (10, 4), f"Coord of wrong shape {coord.shape}"
        assert img.shape == (10, 262144), f"img of wrong shape {img.shape}"

    def test_preprocess_3d_seg(self):
        coord, img = preprocess_data(
            coord=self.coord_dir, size=64, normalization="rescale"
        )
        assert coord.ndim == 2
        assert coord.shape == (10, 4), f"Coord of wrong shape {coord.shape}"
        assert np.all(img == 0), "Image type not zeros"
        assert img.shape == (64, 64, 64), f"img of wrong shape {img.shape}"

    def test_preprocess3d(self):
        coord, _, graph = preprocess_data(
            coord=self.coord_dir, include_label=False, size=64, normalization="rescale"
        )
        assert coord.ndim == 2, "Incorrect no. of dimension"
        assert coord.shape == (10, 3), f"Coord of wrong shape {coord.shape}"
        assert graph.shape == (10, 10), "Graph of wrong shape!"

    def test_filament_mt_dataloader(self):
        train_dl = FilamentDataset(
            coord_dir=join(self.dir_, "filament_mt", "train", "masks"),
            coord_format=".CorrelationLines.am",
            patch_if=500,
        )

        # Build first time
        coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)

        assert len(coords_v) == 1
        assert coords_v[0].shape == (10, 3)
        assert graph_v[0].shape == (10, 10)
        assert output_idx[0].shape == (10,)

        # Load from memory
        coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)

        assert len(coords_v) == 1
        assert coords_v[0].shape == (10, 3)
        assert graph_v[0].shape == (10, 10)
        assert output_idx[0].shape == (10,)

        shutil.rmtree("./temp_train")

    def test_filament_mem_dataloader(self):
        train_dl = FilamentDataset(
            coord_dir=join(self.dir_, "filament_mem", "train", "masks"),
            coord_format=".csv",
            patch_if=500,
        )

        # Build first time
        coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)

        assert len(coords_v) == 1
        assert coords_v[0].shape == (122, 2)
        assert graph_v[0].shape == (122, 122)
        assert output_idx[0].shape == (122,)

        # Load from memory
        coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)

        assert len(coords_v) == 1
        assert coords_v[0].shape == (122, 2)
        assert graph_v[0].shape == (122, 122)
        assert output_idx[0].shape == (122,)

        shutil.rmtree("./temp_train")

    def test_scannet_dataloader(self):
        train_dl = ScannetDataset(
            coord_dir=join(self.dir_, "scannet", "train", "masks"),
            coord_format=".ply",
            patch_if=500,
        )

        # Build first time
        coords_v, _, graph_v, output_idx, clx_idx = train_dl.__getitem__(0)

        assert len(coords_v) == 4
        s = coords_v[0].shape[0]
        assert coords_v[0].shape == (s, 3)
        assert graph_v[0].shape == (s, s)
        assert output_idx[0].shape == (s,)
        assert clx_idx[0].shape[0] == 1

        # Load from memory
        coords_v, _, graph_v, output_idx, clx_idx = train_dl.__getitem__(0)

        assert len(coords_v) == 4
        s = coords_v[0].shape[0]
        assert coords_v[0].shape == (s, 3)
        assert graph_v[0].shape == (s, s)
        assert output_idx[0].shape == (s,)
        assert clx_idx[0].shape[0] == 1

        shutil.rmtree("./temp_train")

    def test_scannet_color_dataloader(self):
        train_dl = ScannetColorDataset(
            coord_dir=join(self.dir_, "scannet", "train", "masks"),
            coord_format=".ply",
            patch_if=500,
        )

        # Build first time
        coords_v, rgb_idx, graph_v, output_idx, clx_idx = train_dl.__getitem__(0)

        assert len(coords_v) == 4
        s = coords_v[0].shape[0]
        assert coords_v[0].shape == (s, 3)
        assert graph_v[0].shape == (s, s)
        assert output_idx[0].shape == (s,)
        assert clx_idx[0].shape[0] == s
        assert rgb_idx[0].shape[0] == s

        # Load from memory
        coords_v, rgb_idx, graph_v, output_idx, clx_idx = train_dl.__getitem__(0)

        assert len(coords_v) == 4
        s = coords_v[0].shape[0]
        assert coords_v[0].shape == (s, 3)
        assert graph_v[0].shape == (s, s)
        assert output_idx[0].shape == (s,)
        assert clx_idx[0].shape[0] == s
        assert rgb_idx[0].shape[0] == s

        shutil.rmtree("./temp_train")

    def test_partnet_dataloader(self):
        train_dl = PartnetDataset(
            coord_dir=join(self.dir_, "partnet", "train", "masks"),
            coord_format=".ply",
            patch_if=500,
        )

        # Build first time
        coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)

        assert len(coords_v) == 4
        s = coords_v[0].shape[0]
        assert coords_v[0].shape == (s, 3)
        assert graph_v[0].shape == (s, s)
        assert output_idx[0].shape == (s,)

        # Load from memory
        coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)

        assert len(coords_v) == 4
        s = coords_v[0].shape[0]
        assert coords_v[0].shape == (s, 3)
        assert graph_v[0].shape == (s, s)
        assert output_idx[0].shape == (s,)

        shutil.rmtree("./temp_train")

    # def test_s3dis_dataloader(self):
    #     train_dl = Stanford3DDataset(
    #         coord_dir=join(self.dir_, "s3dis", "train", "masks"),
    #         coord_format=".ply",
    #         patch_if=500,
    #     )
    #
    #     # Build first time
    #     coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)
    #
    #     assert int(len(coords_v)) == 10
    #     s = coords_v[0].shape[0]
    #     assert coords_v[0].shape == (s, 3)
    #     assert graph_v[0].shape == (s, s)
    #     assert output_idx[0].shape == (s,)
    #
    #     # Load from memory
    #     coords_v, _, graph_v, output_idx, _ = train_dl.__getitem__(0)
    #
    #     assert int(len(coords_v)) == 10
    #     s = coords_v[0].shape[0]
    #     assert coords_v[0].shape == (s, 3)
    #     assert graph_v[0].shape == (s, s)
    #     assert output_idx[0].shape == (s,)
    #
    #     shutil.rmtree("./temp_train")



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from typing import Optional, Union

import torch
import torch.nn as nn
import torch.nn.functional as F


class NodeEmbedding(nn.Module):
    """
    NodeEmbedding class for transforming node features using either a learned
    linear mapping or a randomized cosine transformation, depending on the
    value of sigma.

    This class is used to embed input node features (e.g., RGB values or image
    patches) into a desired output dimension. If sigma is 0, a trainable linear
    layer is applied. Otherwise, a fixed random projection is utilized with
    a cosine activation.
    """

    def __init__(self, n_in: int, n_out: int, sigma=1):
        """
        This class initializes a custom layer with weights and biases. Depending on the
        value of sigma, it either initializes a traditional linear layer without a bias
        or computes weights and biases for a custom mathematical operation. It also
        registers these parameters as buffers for later use if sigma is not zero.

        :param n_in: Number of input features.
        :type n_in: int
        :param n_out: Number of output features.
        :type n_out: int
        :param sigma: Standard deviation used for weight initialization. If zero,
            a standard linear layer is initialized instead. Defaults to 1.
        :type sigma: float
        """
        super().__init__()

        self.linear = None
        if sigma == 0:
            self.linear = nn.Linear(n_in, n_out, bias=False)
        else:
            self.n_in = n_in
            self.sigma = torch.tensor(sigma, dtype=torch.float32)

            w = torch.randn(n_out, n_in)
            b = torch.rand(n_out) * 2 * torch.pi

            self.register_buffer("weight", w)
            self.register_buffer("bias", b)

    def forward(
        self, input_node: Optional[torch.Tensor] = None
    ) -> Optional[torch.Tensor]:
        """
        Performs the forward pass of the module. If an input tensor is provided, it processes
        it through the defined `linear` layer if available, otherwise applies a specific
        transformation involving the cosine function and linear operation.

        :param input_node: Input tensor to process. If None, returns None.
        :type input_node: Optional[torch.Tensor]
        :return: A tensor where the processed input is transformed and scaled into the
            range [0, 1], or None if no input was provided.
        :rtype: Optional[torch.Tensor]
        """
        if input_node is None:
            return None

        if self.linear is not None:
            node = self.linear(input_node)
        else:
            node = torch.cos(F.linear(input_node, self.weight / self.sigma, self.bias))

        return (node + 1) / 2


class EdgeEmbedding(nn.Module):
    """
    EdgeEmbedding layer encapsulates the functionality of computing edge-based
    representations within a graph, leveraging Gaussian radial basis functions (RBF)
    as embedding mechanisms. This module enables the transformation of edge distances
    into either fixed-dimensional encodings or dynamically adjustable encodings
    based on input configurations.

    EdgeEmbedding computes Gaussian kernel representations of edge distances
    and optionally applies a linear transformation to produce embeddings
    of a specified dimensionality. It supports both fixed sigma values (as single
    or iterable ranges) and learns to dynamically adjust their dimensions
    via linear layers.
    """

    def __init__(self, n_out: int, sigma: Union[int, float, list]):
        super().__init__()

        self.n_out = n_out
        self.sigma = sigma

        if isinstance(sigma, (list, tuple)):
            if len(sigma) == 2:
                self._range = torch.linspace(sigma[0], sigma[1], n_out)
            else:
                self._range = torch.linspace(sigma[0], sigma[1], int(sigma[2]))
            assert (
                len(self._range) <= n_out
            ), f"Sigma range is out of shape. n_out = {n_out} but sigma range = {len(self._range)}"

            if len(self._range) != self.n_out:
                self.linear = nn.Linear(len(self._range), n_out, bias=False)
            else:
                self.linear = None
        else:
            self.linear = nn.Linear(1, n_out, bias=False)

    def forward(self, input_coord: torch.Tensor) -> torch.Tensor:
        """
        Computes a transformation of pairwise distances between input coordinates
            and applies an optional linear transformation. The specific transformation
            depends on the configuration, such as whether a fixed `sigma` value or
            varying `_range` values are provided. Handles missing values robustly by
            replacing NaN distance values with zeros.

        :param input_coord: Tensor of shape (..., L, D) where L is the number of
            coordinate points and D is the dimensionality of each point. It represents
            the input feature coordinates for which pairwise distances will be computed.
        :type input_coord: torch.Tensor

        :return: Tensor of shape (..., L, L, K) where K is either 1 or the length of the `_range` attribute.
            Represents the transformed pairwise distances. If `self.linear` is provided, the returned tensor
            is further transformed using the linear layer.
        :rtype: torch.Tensor
        """
        g_len = input_coord.shape[-2]
        g_range = range(g_len)

        dist = torch.cdist(input_coord, input_coord)
        if isinstance(self.sigma, (int, float)):
            dist = torch.exp(-(dist**2) / (self.sigma**2 * 2))

            dist[torch.isnan(dist)] = 0.0
            # isnan = torch.isnan(dist)
            # dist = torch.where(isnan, torch.zeros_like(dist), dist)

            # Overwrite diagonal with 1
            dist = dist.unsqueeze(3)
            dist[:, g_range, g_range, :] = 1.0
        else:
            _range_expanded = self._range.view(1, 1, 1, -1).to(dist.device)
            dist = dist.unsqueeze(-1)
            dist = torch.exp(-(dist**2) / (_range_expanded**2 * 2))
            dist[torch.isnan(dist)] = 0.0
            dist[:, g_range, g_range, :] = 1.0

            # dist_range = torch.zeros(
            #     (1, g_len, g_len, len(self._range)), device=dist.device
            # )
            #
            # for id_1, i in enumerate(self._range):
            #     dist_range[..., id_1] = torch.exp(-(dist**2) / (i**2 * 2))
            #
            # dist_range[torch.isnan(dist_range)] = 0.0
            # dist_range[:, g_range, g_range, :] = 1.0

        if self.linear is not None:
            return self.linear(dist)
        else:
            return dist



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from torch.nn import init

from tardis_em.utils.errors import TardisError


def weights_init_kaiming(m):
    """
    Initializes the weights of layers in a neural network module using the Kaiming
    initialization technique. This function checks the type of layer contained in
    the past module and applies different initialization strategies depending on
    the layer type. Specifically, it adjusts the weights and biases for convolutional
    layers, batch normalization layers, and group normalization layers. This function
    is commonly used for improving the convergence of deep learning models with
    ReLU activation functions.

    :param m: The module whose weights need to be initialized. It is expected
              to be an instance of a layer or a module class such as Conv2d,
              Conv3d, BatchNorm, or GroupNorm.

    :return: None
    """
    class_name = m.__class__.__name__

    if class_name.find("Conv3d") != -1:
        init.kaiming_normal_(tensor=m.weight.data)
    elif class_name.find("Conv2d") != -1:
        init.kaiming_normal_(tensor=m.weight.data)
    elif class_name.find("BatchNorm") != -1:
        init.normal_(tensor=m.weight.data, mean=1.0, std=0.02)
        init.constant_(tensor=m.bias.data, val=0.0)
    elif class_name.find("GroupNorm") != -1:
        init.normal_(tensor=m.weight.data, mean=1.0, std=0.02)
        init.constant_(tensor=m.bias.data, val=0.0)


def init_weights(net, init_type="kaiming"):
    """
    Initializes the weights of a neural network based on the specified initialization
    type. The function applies the kaiming initialization if `init_type` is set to
    "kaiming". If the provided `init_type` is not implemented, an error is raised.

    :param net: The neural network whose weights need to be initialized.
    :param init_type: A string specifying the type of weight initialization to be used.
        Defaults to "kaiming". Must be one of the implemented initialization methods.

    :return: None
    """
    if init_type == "kaiming":
        net.apply(weights_init_kaiming)
    else:
        TardisError(
            "140",
            "tardis_em/cnn/ini_weights.py",
            f"initialization method {init_type} is not implemented!",
        )



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################

from math import pow, sqrt
from typing import Tuple, Union

import numpy as np
from skimage import draw

from tardis_em.cnn.data_processing.interpolation import interpolation
from tardis_em.utils.errors import TardisError


def draw_instances(
    mask_size: Union[list, tuple],
    coordinate: np.ndarray,
    pixel_size: float,
    circle_size=250,
    label=True,
    dtype=None,
) -> np.ndarray:
    """
    Draws labeled or binary masks based on the input coordinates, mask size, and additional parameters.

    This function generates either 2D or 3D masks, depending on the shape of the provided coordinates.
    If label generation is enabled, unique segment labels are created to distinguish between different segments.
    Circles centered around specific points are drawn in the masks, with sizes determined by the input circle size
    and pixel size. Input parameters related to the shape, size, and type of the mask, as well as the labeling behavior,
    are fully customizable. This function is suitable for constructing semantic or instance masks.

    :param mask_size: The dimensions of the mask to be created.
    :type mask_size: list | tuple
    :param coordinate: An array of coordinates specifying the locations to draw the mask.
    :type coordinate: np.ndarray
    :param pixel_size: Size of a pixel in the mask, used to scale the mask appropriately.
    :type pixel_size: float
    :param circle_size: Diameter of the circle to be drawn at each coordinate point. Defaults to 250.
    :type circle_size: int, optional
    :param label: Flag indicating whether a labeled mask or a binary mask should be created. Defaults to True.
    :type label: bool, optional
    :param dtype: Data type for the output mask. If not provided, defaults to np.uint16 or np.uint8.
    :type dtype: str | None, optional

    :return: A 2D or 3D mask generated based on the input parameters.
    :rtype: np.ndarray
    """
    if label:
        if coordinate.ndim != 2 and coordinate.shape[1] not in [3, 4]:
            TardisError(
                "113",
                "tardis_em/cnn/data_processing.md/draw_mask.py",
                "Coordinates are of not correct shape, expected: "
                f"shape [Label x X x Y x (Z)] but {coordinate.shape} given!",
            )

    if label:
        label_mask = np.zeros(mask_size, dtype=np.uint16 if dtype is None else dtype)
    else:
        label_mask = np.zeros(mask_size, dtype=np.uint8 if dtype is None else dtype)

    if pixel_size == 0:
        pixel_size = 1

    r = (circle_size // 2) // pixel_size

    if coordinate.shape[1] == 3 or not label:  # Draw 2D mask
        mask_shape = "c"
    else:  # Draw 3D mask
        mask_shape = "s"

    # Number of segments in coordinates
    if label:
        segments = np.unique(coordinate[:, 0])

        """Build semantic mask by drawing circle in 2D for each coordinate point"""
        for i in segments:
            # Pick coordinates for each segment
            points = coordinate[np.where(coordinate[:, 0] == i)[0]][:, 1:]

            label = interpolation(points)
            all_cz, all_cy, all_cx = [], [], []

            """Draw label"""
            for j in range(len(label)):
                c = label[j, :]  # Point center

                if len(c) == 3:
                    cz, cy, cx = draw_mask(
                        r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                    )
                    all_cz.append(cz)
                    all_cy.append(cy)
                    all_cx.append(cx)
                else:
                    cy, cx = draw_mask(
                        r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                    )
                    all_cy.append(cy)
                    all_cx.append(cx)

            if len(c) == 3:
                all_cz, all_cy, all_cx = (
                    np.concatenate(all_cz),
                    np.concatenate(all_cy),
                    np.concatenate(all_cx),
                )
                label_mask[all_cz, all_cy, all_cx] = i + 1
            else:
                all_cy, all_cx = (
                    np.concatenate(all_cy),
                    np.concatenate(all_cx),
                )
                label_mask[all_cy, all_cx] = i + 1

        return label_mask
    else:
        # Pick coordinates for each point
        all_cz, all_cy, all_cx = [], [], []

        if np.all(coordinate[:, 2] == 0):
            coordinate = coordinate[:, :2]

        for c in coordinate:
            if len(c) == 3:
                cz, cy, cx = draw_mask(
                    r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                )
                all_cz.append(cz)
                all_cy.append(cy)
                all_cx.append(cx)
            else:
                cy, cx = draw_mask(
                    r=r, c=c, label_mask=label_mask, segment_shape=mask_shape
                )
                all_cy.append(cy)
                all_cx.append(cx)
                all_cz.append(np.repeat(0, len(cx)))

        all_cz, all_cy, all_cx = (
            np.concatenate(all_cz),
            np.concatenate(all_cy),
            np.concatenate(all_cx),
        )

        if coordinate.shape[1] == 2:
            label_mask[all_cy, all_cx] = 1
        else:
            label_mask[all_cz, all_cy, all_cx] = 1

        return np.where(label_mask == 1, 1, 0).astype(np.uint8)


def draw_semantic_membrane(
    mask_size: tuple, coordinate: np.ndarray, pixel_size: float, spline_size=70
) -> np.ndarray:
    """
    Draws a semantic membrane mask based on given coordinates, pixel size, and spline size.
    The function generates a mask of the specified size and marks regions centered
    around the provided coordinates. The region size is determined by the given spline size,
    adjusted by the pixel size.

    :param mask_size: The dimensions of the output mask, specified as a tuple of integers (z, y, x).
    :type mask_size: tuple
    :param coordinate: A numpy array of coordinates specifying the center points of each region to be drawn in the mask.
    :type coordinate: np.ndarray
    :param pixel_size: Resolution of the mask, defining the relationship between spline size and the mask grid.
    :type pixel_size: float
    :param spline_size: The diameter of the spline in real-world size, measured in the same unit as pixel_size. Default is 70.
    :type spline_size: int, optional

    :return: A numpy ndarray representing the mask, where regions centered at the provided coordinates are filled.
    :rtype: np.ndarray
    """
    # Ensure ints
    coordinate = coordinate.astype(np.int32)

    # Initiate mask size
    r = round((spline_size / 2) / pixel_size)

    # Initiate mask
    label_mask = np.zeros(mask_size, dtype=np.uint8)

    """Iterate throw each coord"""
    for i in coordinate:
        c = i[1:]

        cz, cy, cx = draw_mask(r=r, c=c, label_mask=label_mask, segment_shape="s")

        label_mask[cz, cy, cx] = 1

    return label_mask


def draw_instances_membrane(
    mask_size: tuple, coordinate: np.ndarray, pixel_size: float, spline_size=70
) -> np.ndarray:
    """
    Draws the instances of membranes on a specified mask by iterating through given
    coordinates, creating circular "stamp"-like regions based on the spline size and
    pixel size, and assigning unique identifiers to each region in the mask.

    The mask is a 3D array where each element corresponds to a region assigned to a
    specific identifier. The process converts coordinates to integers, calculates a
    radius from the given spline and pixel sizes, and applies the circular regions
    to the mask for each set of input coordinates.

    :param mask_size: A tuple representing the size of the mask, typically in 3D, that
        specifies its spatial dimensions.
    :param coordinate: A 2D numpy array where each row corresponds to an instance,
        containing the identifier in the first column and the coordinates in the remaining
        columns.
    :param pixel_size: A float representing the physical distance represented by
        a single pixel in the spatial dimensions of the mask.
    :param spline_size: An integer specifying the base size of the "stamped" mask
        regions. Optional, default value is 70.

    :return: A 3D numpy array with the same dimensions as the input mask, containing
        unique identifiers based on the given coordinate for each "stamped" membrane region.
    :rtype: numpy.ndarray
    """
    # Ensure ints
    coordinate = coordinate.astype(np.int32)

    # Initiate mask size
    r = round((spline_size / 2) / pixel_size)

    # Initiate mask
    label_mask = np.zeros(mask_size, dtype=np.uint16)

    """Iterate throw each coord"""
    for i in coordinate:
        c = i[1:]
        id = i[0]

        cz, cy, cx = draw_mask(r=r, c=c, label_mask=label_mask, segment_shape="s")

        label_mask[cz, cy, cx] = id

    return label_mask


def draw_mask(
    r: int, c: np.ndarray, label_mask: np.ndarray, segment_shape: str
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """
    Draws a mask on the given label array by creating a circle or sphere depending
    on the specified shape. The method can handle both 2D and 3D label masks.

    :param r: The radius of the circle (2D) or sphere (3D) to be drawn.
    :param c: A numpy array containing the center coordinates of the shape,
        where the length of the input determines if it is 2D or 3D.
    :param label_mask: A numpy array representing the label mask where
        the circle or sphere will be drawn. Must have 2 or 3 dimensions.
    :param segment_shape: A string specifying the shape to draw; `"c"` for
        circle (default) or `"s"` for sphere.

    :return: A tuple of numpy arrays representing the coordinates of the
        drawn mask shape. For 3D shapes, the tuple will include `z`, `y`,
        and `x` arrays. For 2D shapes, the tuple will include `y` and `x`
        arrays, omitting the `z` array.
    """
    if label_mask.ndim not in [2, 3]:
        TardisError(
            "113",
            "tardis_em/cnn/data_processing.md/draw_mask.py"
            f"Unsupported dimensions given {label_mask.ndim} expected 2!",
        )

    x = int(c[0])
    y = int(c[1])
    if len(c) == 3:
        z = int(c[2])
    else:
        z = None

    if segment_shape not in ["s", "c"]:
        TardisError(
            "123",
            "tardis_em/cnn/data_processing.md/draw_mask.py"
            f"Unsupported shape type to draw given {segment_shape} but "
            'expected "c" - circle or "s" - sphere',
        )
    if segment_shape == "s":
        cz, cy, cx = draw_sphere(r=r, c=(z, y, x), shape=label_mask.shape)
        return cz, cy, cx
    else:
        if z is not None:
            cz, cy, cx = draw_circle(r=r, c=(z, y, x), shape=label_mask.shape)
            return cz, cy, cx
        else:
            cy, cx = draw_circle(r=r, c=(y, x), shape=label_mask.shape)
            return cy, cx


def draw_circle(
    r: int, c: tuple, shape: tuple
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """
    Generates the coordinates of pixels to draw a filled circle in a given array shape.
    Depending on the input `c` (2D or 3D), it generates either 2D (y, x) or 3D (z, y, x)
    coordinates restricted within the specified shape.

    The method uses the `skimage.draw.disk` function to calculate the circular region
    and adjusts for its position in the provided coordinate frame `c`. It ensures
    that all pixels remain within the bounds of the array defined by `shape`.

    :param r: Radius of the circle to be drawn.
    :param c: Tuple representing the center coordinates of the circle. In 2D, it requires (y, x),
        and for 3D, it requires (z, y, x).
    :param shape: Tuple representing the shape of the frame or array where the circle needs to be
        drawn. For 2D: (rows, columns). For 3D: (depth, rows, columns).

    :return: Coordinates of pixels forming the circle. A tuple of arrays (y, x) for 2D or
        (z, y, x) for 3D representing the indices of the circle's pixels.
    """
    r_dim = round(r * 3)
    c_frame = round(r)

    ny, nx = draw.disk((r, r), r, shape=(r_dim, r_dim))
    if len(c) == 3:
        nz = np.repeat(c[0], len(nx))

        c = ((c[1] - c_frame), (c[2] - c_frame))
        z, y, x = nz, ny + c[0], nx + c[1]

        # Remove pixel out of frame
        zyx = np.array((z, y, x)).T
        del_id = []
        for id_, i in enumerate(zyx):
            if i[0] >= shape[0] or i[1] >= shape[1] or i[2] >= shape[2]:
                del_id.append(id_)

        zyx = np.delete(zyx, del_id, 0)

        return zyx[:, 0], zyx[:, 1], zyx[:, 2]
    else:
        c = ((c[0] - c_frame), (c[1] - c_frame))
        y, x = ny + c[0], nx + c[1]

        # Remove pixel out of frame
        yx = np.array((y, x)).T
        del_id = []
        for id_, i in enumerate(yx):
            if i[0] >= shape[0] or i[1] >= shape[1]:
                del_id.append(id_)

        yx = np.delete(yx, del_id, 0)

        return yx[:, 0], yx[:, 1]


def draw_sphere(
    r: int, c: tuple, shape: tuple
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Generates the coordinates of a 3D spherical structure within a volume of specified shape.
    This function simulates a sphere using a 3D binary array, applies trimming to the sphere,
    and shifts its position within a given coordinate frame. The output consists of the
    adjusted coordinates of the sphere within the specified 3D volume.

    :param r: Radius of the sphere
    :type r: int
    :param c: Center coordinates of the sphere in the 3D volume
    :type c: tuple
    :param shape: Shape of the 3D volume to contain the sphere
    :type shape: tuple

    :return: Coordinates of the sphere within the 3D volume along z, y, and x axes
    :rtype: tuple of numpy.ndarray
    """
    r_dim = round(r * 2)
    sphere_frame = np.zeros((r_dim, r_dim, r_dim), dtype=np.int8)
    trim = round(r_dim / 6)
    c_frame = round(r)

    z, y, x = sphere_frame.shape
    for z_dim in range(z):
        for y_dim in range(y):
            for x_dim in range(x):
                dist_zyx_to_c = sqrt(
                    pow(abs(z_dim - c_frame), 2)
                    + pow(abs(y_dim - c_frame), 2)
                    + pow(abs(x_dim - c_frame), 2)
                )
                if dist_zyx_to_c > c_frame:
                    sphere_frame[z_dim, y_dim, x_dim] = False
                else:
                    sphere_frame[z_dim, y_dim, x_dim] = True
    sphere_frame[:trim, :] = False  # Trim bottom of the sphere
    sphere_frame[-trim:, :] = False  # Trim top of the sphere
    z, y, x = np.where(sphere_frame)

    c = ((c[0] - c_frame), (c[1] - c_frame), (c[2] - c_frame))
    z, y, x = z + c[0], y + c[1], x + c[2]

    # Remove pixel out of frame
    zyx = np.array((z, y, x)).T
    del_id = []
    for id_, i in enumerate(zyx):
        if i[0] >= shape[0] or i[1] >= shape[1] or i[2] >= shape[2]:
            del_id.append(id_)
        if i[0] < 0:  # Remove negative value
            del_id.append(id_)

    zyx = np.delete(zyx, del_id, 0)

    return zyx[:, 0], zyx[:, 1], zyx[:, 2]



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2025                                            #
#######################################################################
from os import listdir, mkdir
from os.path import isdir
from shutil import rmtree

import numpy as np

from tardis_em.cnn.data_processing.stitch import StitchImages
from tardis_em.cnn.data_processing.trim import trim_with_stride


def test_trim_stitch_3d():
    stitch = StitchImages()

    if isdir("./tests/test_data/temp"):
        rmtree("./tests/test_data/temp")
    mkdir("./tests/test_data/temp")

    image_3d_15 = np.zeros((15, 15, 15), dtype=np.float32)
    mask_3d_15 = np.zeros((15, 15, 15), dtype=np.uint8)
    image_3d_150 = np.zeros((150, 150, 150), dtype=np.float32)
    mask_3d_150 = np.zeros((150, 150, 150), dtype=np.uint8)

    trim_with_stride(
        image=image_3d_15,
        trim_size_xy=15,
        trim_size_z=15,
        output="./tests/test_data/temp",
        image_counter=0,
        scale=(15, 15, 15),
        clean_empty=False,
        mask=mask_3d_15,
    )
    assert len(listdir("./tests/test_data/temp/imgs")) == 1
    assert len(listdir("./tests/test_data/temp/masks")) == 1

    stitch_3d_15 = stitch(
        image_dir="./tests/test_data/temp/imgs", mask=False, dtype=np.float32
    )
    assert stitch_3d_15.shape == (15, 15, 15)

    stitch_3d_15 = stitch(
        image_dir="./tests/test_data/temp/masks",
        mask=True,
        prefix="_mask",
        dtype=np.uint8,
    )
    assert stitch_3d_15.shape == (15, 15, 15)

    rmtree("./tests/test_data/temp")
    mkdir("./tests/test_data/temp")

    trim_with_stride(
        image=image_3d_150,
        trim_size_xy=64,
        trim_size_z=64,
        output="./tests/test_data/temp",
        image_counter=0,
        stride=0,
        scale=(150, 150, 150),
        clean_empty=False,
        mask=mask_3d_150,
    )
    assert len(listdir("./tests/test_data/temp/imgs")) == 27
    assert len(listdir("./tests/test_data/temp/masks")) == 27

    stitch_3d_15 = stitch(
        image_dir="./tests/test_data/temp/imgs", mask=False, dtype=np.float32
    )
    assert stitch_3d_15.shape == (192, 192, 192)

    stitch_3d_15 = stitch(
        image_dir="./tests/test_data/temp/masks",
        mask=True,
        prefix="_mask",
        dtype=np.uint8,
    )
    assert stitch_3d_15.shape == (192, 192, 192)

    rmtree("./tests/test_data/temp")


def test_trim_stitch_2d():
    stitch = StitchImages()

    if isdir("./tests/test_data/temp"):
        rmtree("./tests/test_data/temp")
    mkdir("./tests/test_data/temp")

    image_2d_15 = np.zeros((15, 15), dtype=np.float32)
    mask_2d_15 = np.zeros((15, 15), dtype=np.uint8)
    image_2d_150 = np.zeros((150, 150), dtype=np.float32)
    mask_2d_150 = np.zeros((150, 150), dtype=np.uint8)

    trim_with_stride(
        image=image_2d_15,
        trim_size_xy=15,
        trim_size_z=15,
        output="./tests/test_data/temp",
        image_counter=0,
        scale=(15, 15),
        clean_empty=False,
        mask=mask_2d_15,
    )
    assert len(listdir("./tests/test_data/temp/imgs")) == 1
    assert len(listdir("./tests/test_data/temp/masks")) == 1

    stitch_2d_15 = stitch(
        image_dir="./tests/test_data/temp/imgs", mask=False, dtype=np.float32
    )
    assert stitch_2d_15.shape == (15, 15)

    stitch_2d_15 = stitch(
        image_dir="./tests/test_data/temp/masks",
        mask=True,
        prefix="_mask",
        dtype=np.uint8,
    )
    assert stitch_2d_15.shape == (15, 15)

    rmtree("./tests/test_data/temp")
    mkdir("./tests/test_data/temp")

    trim_with_stride(
        image=image_2d_150,
        trim_size_xy=64,
        trim_size_z=64,
        output="./tests/test_data/temp",
        image_counter=0,
        stride=0,
        scale=(150, 150),
        clean_empty=False,
        mask=mask_2d_150,
    )
    assert len(listdir("./tests/test_data/temp/imgs")) == 9
    assert len(listdir("./tests/test_data/temp/masks")) == 9

    stitch_2d_15 = stitch(
        image_dir="./tests/test_data/temp/imgs", mask=False, dtype=np.float32
    )
    assert stitch_2d_15.shape == (192, 192)
    stitch_2d_15 = stitch(
        image_dir="./tests/test_data/temp/masks",
        mask=True,
        prefix="_mask",
        dtype=np.uint8,
    )
    assert stitch_2d_15.shape == (192, 192)

    rmtree("./tests/test_data/temp")
